#!/usr/bin/env python

'''
AMiGA library of functions for parsing and organizing the working directory.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (9 functions)

# validateDirectories
# checkDirectoryExists
# checkDirectoryNotEmpty
# printDirectoryContents
# mapDirectories
# mapFiles
# isFileOrFolder
# assemblePath
# assembleFileName

import os
import sys


def validateDirectories(directory,verbose=False):
    '''
    Checks specific directories for their existence and makes sure data was 
        provided by user. It will also compose and can print a message that can
        communicate with user the results of this validation.

    Args:
        directory (dict): dictionary where keys are AMiGA-relevant folders and 
            values are the corresponding file paths
        verbose (boolean): if True print message
 
    Returns:
        None
    '''

    # params (dict) defines parameters for directory validation
    #     
    #     key: directory in directory-storing dictionary (str) 
    #
    #     value (4-tuples) where:    
    #     1. generic name of directory for communcating with user (str)
    #     2. initialization: whether to initialize directory or not (boolean)
    #     3. force system exit: whether to exist system if error detected

    params = {
        'parent':('Input directory',False,True),
        'data':('Data directory',False,False),
        'derived':('Derived data directory',True,False),
        'mapping':('Mapping directory',True,False),
        'summary':('Summary directory',True,False),
        'figures':('Figures directory',True,False),
        'models':('Models directory',True,False)
    }

    full_msg = ''

    # check whether these directories exist and/or take action based on parameters 
    for i,params_i in params.items():
        _,msg = checkDirectoryExists(directory[i],*params_i)
        full_msg += msg

    # check whether user provided at least one data file
    _,msg = checkDirectoryNotEmpty(directory['data'],params['data'][0])
 
    full_msg += '\n' # message will multi-line and bottom is padded with a new line
    full_msg += msg # itemized list of content of data folder

    if verbose:
        print(full_msg)

    return None


def checkDirectoryExists(directory,generic_name='directory',
                         initialize=False,sys_exit=False,max_width=25):
    '''
    Returns a logic argument (True=exists) and a message that could be relayed 
        to user in regard to existence status and actions taken.
    
    Args:
        directory (str): path
        generic_name (str): short name of directory (for communication with user only)
        sys_exit (boolean): request premature termination if directory does not exist
        initialize (boolean): if directory does not exist, initialize it
        max_width (int): used for padding width in string formatting  

    Returns:
        arg (boolean): True indicates that directory exists
        msg (str):  message composed for communication with user
    '''

    # check if directory exists
    exists = os.path.exists(directory)

    # perform action based on whether directory existss and input parameters
    if exists:
        arg = True
        msg = '{:.<{width}}{}\n'.format(generic_name,directory,width=max_width)
    elif sys_exit:
        arg = False
        sys.exit(f'FATAL USER ERROR: {generic_name} {directory} does not exist.\n')
    elif initialize:
        os.makedirs(directory)
        arg = True
        msg = f'WARNING: {directory} did not exist but was created.\n'
    else:
        arg = False
        msg = f'WARNING: {directory} does not exist.\n'

    return arg,msg


def checkDirectoryNotEmpty(directory,generic_name='Directory'):
    '''
    Checks if a directory has at least one file.

    Args:
        directory (str): path
        generic_name (str): generic name of directory for communcating with user

    Returns:
        arg (boolean): True indicates that directory exists
        msg (str):  message composed for communication with user
    '''

    numFiles = len(os.listdir(directory))

    if numFiles == 0:
        arg = False
        sys.exit(f'FATAL USER ERROR: {generic_name} {directory} is empty.\n')
    else:
        arg = True
        msg = f'{generic_name} {directory} has {numFiles} files:'
        msg += '\n\n'
        msg += '\n'.join(printDirectoryContents(directory)) # print one item per line
        msg += '\n' # pad bottom withe new line

    return arg,msg


def printDirectoryContents(directory,sort=True,tab=True):
    '''
    prints a list of contents in a directory

    Args:
        directory (str): path
        sort (boolean)
        tab (boolean): pads each item with (4 spaces) in front

    Returns:
        items (list) where elements are file names (str)
    '''

    items = os.listdir(directory)
    
    if sort:
        items = sorted(items)

    if tab:
        items = [f'    {i}' for i in items]

    return items


def mapDirectories(parent):
    '''
    Returns a dictionary where keys are AMiGA-relevant folders and 
        values are the corresponding file paths

    Args:
        parent (str): file path

    Returns:
        directory (dict) where keys and values are strings
    '''

    # initialize dictionary where values are file paths
    directory = {}

    # topmost directory is the input parent directory
    directory['parent'] = parent

    # format file paths for sub-directories 
    children = ['data','derived','mapping','summary','parameters','figures','models']

    for child in children:

        directory[child] = assemblePath(parent,child)

    return directory


def mapFiles(directory):
    '''
    Returns a dictionary where keys are AMiGA-relevant files and
        vlaues are the corresponding file paths
    '''

    files  = {}

    # only single file of interest in 'mapping' sub-directory
    files['meta'] = '{}{}{}.txt'.format(directory['mapping'],os.sep,'meta')

    # format paths for files in the 'parameter' sub-directory
    children = ['flag','hypothesis','subset','interval']

    for child in children:

        files[child] = assemblePath(directory['parameters'],child,'.txt')

    return files


def isFileOrFolder(filepath,up=2):
    '''
    Determines if a path points to a file or directory.

    Args:
        filepath (str): file path 
        up (int): which directoy to return if a file, the parent (1) or grandparent (2)
    '''

    isFile = os.path.isfile(filepath)

    if isFile:
        if up == 2:
            parent = os.path.dirname(os.path.dirname(filepath))
        elif up == 1:
            parent = os.path.dirname(filepath)
        filename = os.path.basename(filepath)
        if parent == '' or parent is None:
            parent = '.'
        return parent,filename
    else:
        parent = filepath
        return parent,None


def assemblePath(directory,filebase,extension=''):
    '''
    Stitches a full path to a file.

    Args:
        directory (str): path to directory
        filename (str): file name with extension
        extension (str): by default '', allows stitching filenames that alreay have extension

    Returns:
        filepath (str)
    '''

    # check if directory ends with slash, used to avoid double slashes ('//')
    sep = ['' if directory[-1]==os.sep else os.sep][0]
    file_path = f'{directory}{sep}{filebase}{extension}'

    return file_path


def assembleFullName(folder,prefix,filename,suffix,extension):
    '''
    Assembles a file name using the arguments into a complete full path. 

    Args:
        folder (str): full path to file
        prefix (str): prefix to filename (precedes underscore)
        filename (str): filename without extension
        suffix (str): suffix to filename (succedes underscore)
        extension (str): file extension (e.g. .txt,.tsv,.pdf). should include period. 

    Returns:
        file_path (str): full path to generated file name
    '''

    file_name = f'{prefix}_{filename}_{suffix}'
    file_name = file_name.strip('_') # in case prefix is empty

    file_path = assemblePath(folder,file_name,extension)

    return file_path




#!/usr/bin/env python

'''
AMiGA wrapper: normalizing parameters in summary files.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (3 functions)

# main
# read
# normalize_parameters


import os
import sys
import operator
import pandas as pd # type: ignore

from .comm import smartPrint, tidyMessage
from .interface import checkParameterCommand
from .org import isFileOrFolder 
from .read import findPlateReaderFiles
from .params import initParamList
from .utils import subsetDf


def main(args):

	verbose = args.verbose
	directory = args.input
	ovewrrite = args.over_write

	msg = 'AMiGA is parsing your file(s)'

	smartPrint('',verbose)
	smartPrint(tidyMessage(msg),verbose)
	
	directory,filename = isFileOrFolder(directory,up=1)

	if filename:
		ls_files = [f'{directory}{os.sep}{filename}']
	else:
		ls_files = findPlateReaderFiles(directory,'.txt')

	for lf in ls_files:

		df = read(ls_files)

		if ovewrrite: 
			new_name = lf
		elif lf.endswith('.txt'):
			new_name = f'{lf[:-4]}_normalized.txt'
		else:
			new_name = f'{lf}.normalized.txt'
		
		df = normalizeParameters(args,df)
		df.to_csv(new_name,sep='\t',header=True,index=True)

	msg = 'AMiGA compelted your request'

	smartPrint('',verbose)
	smartPrint(tidyMessage(msg),verbose)


def read(ls_fpaths):
	'''
	Reads all files listed in input argument and concats tables
	    into one large pandas.DataFrame.

	Args:
	    ls_fpaths (list)

	Return:
	    pandas.DataFrame
	'''

	df_list = []

	for fpath in ls_fpaths:
		df_list.append(pd.read_csv(fpath,sep='\t',header=0,index_col=0))

	df = pd.concat(df_list,sort=False)

	return df


def normalizeParameters(args,df):
	'''
	Normalizes growth parameters to control samples. 

	Args:
	    args (dictionary): keys are arguments and value are user/default choices
	    df (pandas.DataFrame): rows are samples, columns are experimental variables. Must include
	        Plate_ID, Group, Control, auc, k, gr, dr, td, lag.

	Returns:
	    df (pandas.DataFrame): input but with an additional 6 columns.
	'''

	# let's keep original dataframe
	df_orig = df.copy()
	df_orig_keys = df_orig.columns 
	df = df.reset_index()

	# How you should normalize?
	if args.normalize_method == 'division':
		opr = operator.truediv
	elif args.normalize_method == 'subtraction':
		opr = operator.sub

	# How to group samples and which ones are control samples?

	# if user specifies with command-line arguments
	if args.group_by is not None and args.normalize_by is not None:

		groupby = args.group_by.split(',')
		controlby = checkParameterCommand(args.normalize_by)

	elif args.normalize_by is not None and args.group_by is None:

		controlby = checkParameterCommand(args.normalize_by)
		df.loc[:,'Group'] = [1]*df.shape[0]
		groupby = ['Group']

	# else check columns for Group and Contol variables
	elif 'Group' in df_orig_keys and 'Control' in df_orig_keys:

		groupby = ['Group']
		controlby = {'Control':1}

		if (len(df.Group.unique())==1) and (len(df.Plate_ID.unique())>1):
			msg = '\nUSER WARNING: AMiGA detected a single "Group" but multiple Plate_IDs.\n'
			msg += 'Wells from different plates will thus be normalized togther as a group.\n'
			msg += 'If this was not your intention, please pass explicit arguments to AMiGA\n'
			msg += 'using "--group-by" and "--control-by" arguments to avoid any ambiguity.\n' 
			print(msg)

    # else exit with error message
	else:
		msg = 'FATAL USER ERROR: User must specify groups of samples and '
		msg += 'their corresponding control samples.'
		sys.exit(msg)

	# which parameters to normalize and/or to keep
	params_1 = initParamList(0)
	params_1.remove('diauxie')
	params_2 = [f'mean({ii})' for ii in params_1]
	params_3 = initParamList(2)

	if any([ii in df_orig_keys for ii in params_2]):  
		params = params_2
	elif any([ii in df_orig_keys for ii in params_3]):
		params = params_3 
	else:
		params = params_1

	#params_norm = initParamList(2)

	params_keep = groupby + list(controlby.keys()) + ['Sample_ID','Plate_ID'] + params
	params_keep = list(df.columns[df.columns.isin(params_keep)])
	params_varbs = list(set(params_keep).difference(set(params)))
	df = df.loc[:,params_keep]

	norm_df = []
	for idx,row in df.loc[:,groupby].drop_duplicates().iterrows():

		df_group = subsetDf(df,row.to_dict()).loc[:,params_keep]
		df_group = df_group.sort_values(params_varbs)
		df_control = subsetDf(df_group,controlby)

		df_group.set_index(params_varbs,inplace=True)
		df_control.set_index(params_varbs,inplace=True)

		dgv = df_group.values
		dcv = df_control.mean(numeric_only=True).values

		df_group.loc[:,:] = opr(dgv,dcv)
		norm_df.append(df_group)	

	norm_df = pd.concat(norm_df,axis=0)
	norm_df.columns = [f'norm({ii})' for ii in norm_df.columns]	
	norm_df = norm_df.reset_index(drop=False)

	df = pd.merge(df_orig,norm_df,on=params_varbs)

	if 'Sample_ID' in df.columns:
		df = df.set_index('Sample_ID')

	return df




#!/usr/bin/env python

'''
AMiGA library of functions for parsing commands and arguments.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (5 functions)

# interpretParameters
# initializeParameter
# checkParameterCommand
# checkParameterText
# integerizeDictValues

import os
import re

from .comm import smartPrint, tidyDictPrint


def interpretParameters(files,args,verbose=False):
    '''
    Checks specific directories for their existence and makes sure data was 
        provided by user. It will also compose and can print a message that can
        communicate with user the results of this validation.

    Args:
        files (dict): keys are parameter names and values are file paths
        args (dict): keys are parameter names and values are corresponding command-line arguments
 
    Returns:
        params_dict (dict): dictionary where keys are variables and values instances
    '''

    # params_settings defines parameters for interpreting parameter files or argument calls
    #     
    #     each item is a 3-tuple where:    
    #     1. parameter name which matches both input argument and text file name (str)
    #     2. delimitor that separates values of a variable of the parameter (str)
    #     3. integerize: whether to convert values of a variable to integers (boolean)

    params_settings = [
        ('interval',',',True),
        ('subset',',',False),
        ('flag', ',',False),
        ('hypothesis',r'\+|,',False)
    ]
    
    # initialize all parameters based on their settings
    params_dict = {}
    for pp,sep,integerize in params_settings:
        params_dict[pp] = initializeParameter(files[pp],getattr(args, pp),sep=sep,integerize=integerize)

    smartPrint(tidyDictPrint(params_dict),verbose)

    # if user requests any subsetting, summary results must be merged
    if params_dict['subset'] and (not args.merge_summary):

        args.merges = True

        msg = 'WARNING: Because user has requested subsetting of data, '
        msg += 'results will be merged into single summary and/or data file.\n'
        smartPrint(msg,verbose)

    return params_dict,args


def initializeParameter(filepath,arg,sep=',',integerize=False):
    '''
    Parses command-line argument and/or corresponding text file to define parameter.

    Args:
        filepath (str): path to parameter text file
        arg (str): the command-line user-passed value to argument/parameter
        sep (str): separator or delimitor used in argument key-values pairs
        integerize (boolean): whether to convert argument values into integers

    Retruns:
        param_dict (dict): dictionary where keys are variables and values instances
    '''

    # if user did not provide argument, check for text file
    if arg is None:
        param_dict = checkParameterText(filepath,sep=sep)

    # else if user provided argument, parse argument for parameters
    elif len(arg)>0 and arg.isdigit():
        param_dict = float(arg)

    elif len(arg)>0:
        param_dict = checkParameterCommand(arg,sep=sep)

    # otherwise, initialize empty dictionary
    else:
        param_dict = {}

    # if argument parameters should be converted to integers (e.g. time interval)
    if integerize and isinstance(param_dict,dict): 
        return integerizeDictValues(param_dict)
    else: 
        return param_dict

def get_sep(sep):
        '''
        Ensures safe handling of non-regex separators by re.split()

        Args:
            sep (str): should be either ',' or '\+|,'
        
        Returns:
            (str)
        '''
        return sep if '|' in sep else re.escape(sep)

def checkParameterCommand(command,sep=','):
    '''
    Parses command-line text argument and formats it into a dictionary.
        (1) Text must be a list of at least one item that are separated by 
        semicolons (;). (2) Each item must be a variable name separated from a 
        list of its values with a colon (:). (3) Each list must have at least
        one value separated by commas (,).

    Args:
        command (str): see above for description of required format
        sep (str): delimitor that separates values of a variable

    Returns:
        lines_dict (dict): keys are variables and values are a list of variable instances
    '''

    if command is None:
        return None

    # strip flanking semicolons and whitespaces then split by semicolons 
    lines = command.strip('; ').split(';')

    # get names of variables (left of semicolons)
    lines_keys = [ii.split(':')[0].strip() for ii in lines]

    # get list of valuse or instances for all variables
    lines_values = [[jj.strip() for jj in re.split(get_sep(sep), ii.split(':')[1])] for ii in lines]

    # re-package variables and their list of values or instances into a dictionary
    lines_dict = {ii:jj for ii,jj in zip(lines_keys,lines_values)}

    return lines_dict


def checkParameterText(filepath,sep=','):
    '''
    Parses text-based parameter file and formats its content into a dictionary.
        (1) Items must be separated by newlines (\n). (2) Each item must be a 
        variable name separated from a list of its values with a colon':'. (3) 
        Each list must have at least one value separated by commas (,).

    Args:
        sep (str): delimitor that separates values of a variable

    Returns:
        lines_dict (dict): keys are variables and values are a list of variable instances
    '''

    exists = os.path.exists(filepath)

    lines_dict = {}

    if exists:
        fid = open(filepath)
        for line in fid:
            key,value = line.split(':')
            values = re.split(get_sep(sep), value.strip('\n'))
            values = [ii.strip() for ii in values]
            values = [float(ii) if ii.isdigit() else ii for ii in values]
            lines_dict[key.strip()] = values

    return lines_dict 


def integerizeDictValues(dictionary):
    '''
    Converts items in the values of a dictionary into integers. This will work for values
        that are iterable (e.g. list).

    Args:
        dictionary (dict)

    Returns:
        dicionary (dict) where each value is a list of integers

    Example: input {'CD630_PM1-1':str(500)} will return {'CD630_PM1-1':int(500)}
    '''

    if dictionary is None:
        return None

    for key,value in dictionary.items():

        if isinstance(value,list):
            dictionary[key] = float(value[0])
        else:
            dictionary[key] = float(value)

    return dictionary




#!/usr/bin/env python

'''
AMiGA library for the GrowthPlate class for storing and manipulting growth curve data.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (1 class with 16 sub-functions)

# GrowthPlate (CLASS)
#   __init__
#   computeBasicSummary
#   subtractControl
#   thinMeasurements
#   computeFoldChange
#   convertTimeUnits
#   logData
#   subtractBaseline
#   isSingleMultiWellPlate
#   saveKey
#   plot
#   addLocation
#   extractGrowthData
#   copy
#   compute_k_error
#   model

import sys
import numpy as np # type: ignore
import pandas as pd # type: ignore
import matplotlib.pyplot as plt # type: ignore
import seaborn as sns # type: ignore

from copy import deepcopy

from .params import initParamDf, mergeDiauxieDfs
from .plot import largeTickLabels
from .model import GrowthModel
from .comm import smartPrint
from .detail import parseWellLayout
from .utils import subsetDf, handle_non_pos, getPlotColors, getTextColors, getValue, getTimeUnits

pd.set_option('future.no_silent_downcasting', True)

class GrowthPlate:

    def __init__(self,data=None,key=None,time=None,input_time=None,input_data=None,mods=None,gp_data=None):
        '''
        Data structure for handling growth data of microtiter plate assays. This however 
            can be generalized to any depenent observations with a single independent observation
            (e.g. time).

        Attributes:

            data (pandas.DataFarme): n x (p+1) DataFrmae (for n timepoints by p wells) that stores
                the raw optical density (or absorbance) measurements. The first column is assumed 
                to be Time. Each succeeding column corresponds to observations for a specific sample
                (i.e. well).
            key (pandas.DataFrame): p x k) DataFrame (for k experimental variables or descriptors) 
                that storees the experimental variables for each sample. One of the columns
                is a unique sample (i.e. well) identifier ('Sample_ID'). You can use Well Locations 
                (e.g. A1) here but if a GrowthPlate object can combine data from multiple plates, 
                thus some wells might share the same well location.
            time (pandas.DataFrame): t x 1 DataFrame (for n timepoints) that stores the raw time measurments.
            input_time (pandas.DataFrame): time attribute but not subjected to any modifications (see mods)
            input_data (pandas.DataFrame): data attribute but not subject to any modifiecations (see mods)
            mods (pandas.DataFrame): tracks which modifications were applied to data & time attributes
            gp_data (pandas.DataFrame): each row is a specific time-point for a specific sample. Identifying columns
                include Plate_ID, Time, Sample_ID. Data columns include
                    + OD (original data)
                    + OD_Fit (gp model fit of original data)
                    + OD_Derivative (gp model fit of derivative, insensitive to y-value, i.e. whether OD is centered)
                    + GP_Input (input to gp.GP() object), this is usually log-transformed and log-baseline-subtracted
                    + GP_Output (output of gp.GP().predict()), hence also log-trasnformed and log-baseline-subtracted
                    + OD_Growth_Data (GP_Input but converted to real OD and centered at zero)
                    + OD_Growth_Fit (GP_Output but converted to real OD and centered at zero)
        '''

        if time is None:
            self.data = data.iloc[:,1:]
            self.time = pd.DataFrame(data.iloc[:,0])
        else:
            self.data = data.copy()
            self.time = time.copy()

        if input_time is None:
            self.input_time = self.time.copy()
        else:
            self.input_time = input_time
        
        if input_data is None:
            self.input_data = self.data.copy()
        else:
            self.input_data = input_data

        self.key = key.copy()
        self.gp_data = gp_data

        if mods is None:
            self.mods = pd.DataFrame(columns=['logged','floored','controlled','raised'],index=['status'])
            self.mods = self.mods.apply(lambda x: False) 
        else:
            self.mods = mods

        assert isinstance(key,pd.DataFrame), "key must be a pandas DataFrame"
        assert (self.data.shape[1]) == (self.key.shape[0]), "key must contain mapping data for all samples"


    def computeBasicSummary(self):
        '''
        Computes basic characteristics of each column in the object's input_data attribute. In particular,
            + OD_Baseline is the first measurment (0-row)
            + OD_Max is the maximum measurment (any row)
            + OD_Min is the miimum measurement (any row)
        '''

        # compute OD_Min, OD_Max, and OD_Baseline
        df = self.input_data.copy()  # timepoints by wells, input data that remains unmodified
        df_max = df.max(axis=0,numeric_only=True)  # max by column (i.e. well)
        df_min = df.min(axis=0,numeric_only=True)  # min by column (i.e. well)
        df_baseline = df.iloc[0,:]

        # compute OD_Emp_AUC
        time = self.time.values
        dt = np.mean(time[1:,0]-time[:-1,0])
        df_auc = df.apply(lambda x: np.dot(np.repeat(dt,df.shape[0]).T,x),axis=0)

        joint_df = pd.concat([df_baseline,df_min,df_max,df_auc],axis=1)
        joint_df.columns = ['OD_Baseline','OD_Min','OD_Max','OD_Emp_AUC']

        self.key = self.key.join(joint_df)


    def subtractControl(self,to_do=False,drop=True,blank=False):
        '''
        Subtract from each treatment sample's growth curve, the growth curve of its corresponding control sample.

        Args:
            to_do (boolean): if False, do not subtract control wells and return None.
            drop (boolean): if True, drop control samples from data.
        '''

        if not to_do:
            return None

        data = self.data.copy()
        mapping = self.key.copy()

        pid_text = 'Plate_ID'
        grp_text = 'Group'
        ctr_text = 'Control'
        if blank:
            grp_text, ctr_text = f'Blank{grp_text}', f'Blank{ctr_text}'

        # find all unique groups
        plate_groups = mapping.loc[:,[pid_text,grp_text]].drop_duplicates()
        plate_groups = [tuple(x) for x in plate_groups.values]

        for plate_group in plate_groups:

            pid,group = plate_group

            # grab lists of Sample_ID of wells corresponding to control and cases
            controls = subsetDf(mapping,{pid_text:[pid],grp_text:[group],ctr_text:[1],'Flag':[0]}).index.values
            cases = subsetDf(mapping,{pid_text:[pid],grp_text:[group]}).index.values  # includes controls

            if len(controls)==0:
                msg = '\nFATAL ERROR: User requested subtraction of control samples. However, '
                msg+= f'samples belonging to group {group} of plate {pid} lack ' 
                msg+= 'any corresponding control samples in the current working directory.\n'
                sys.exit(msg)

            data_controls = data.loc[:,controls]
            data_cases = data.loc[:,cases]

            # for each case, divide data by mean controls (if log-transformed), o.w. subtract mean controls
            data_controls = data_controls.mean(axis=1,numeric_only=True)
            data_cases = (data_cases.T - data_controls).T
            data.loc[:,cases] = data_cases.values

            mapping.loc[cases,'Adj_OD_Baseline'] = data.loc[:,cases].loc[0,:].values
            mapping.loc[cases,'Adj_OD_Min'] = data.loc[:,cases].apply(np.min).values
            mapping.loc[cases,'Adj_OD_Max'] = data.loc[:,cases].apply(np.max).values

            if drop: 
                data = data.drop(controls,axis=1)
                mapping = mapping.drop(controls,axis=0)

        self.data = data
        self.key = mapping
        self.mods.controlled = True


    def thinMeasurements(self,step=11):
        '''
        Thin the number of measurements (for both object's time and data).

        Args:
            step (int)
        '''

        matrix = (self.time).join(self.data)
        select = np.arange(0,self.data.shape[0],step)
        matrix = matrix.iloc[select,:]

        self.time = pd.DataFrame(matrix.iloc[:,0])
        self.data = matrix.iloc[:,1:] 


    def dropFlaggedWells(self,to_do=False):

        if not to_do:
            return None

        flagged = self.key[self.key.Flag==1].index.values

        self.key.drop(labels=flagged,axis=0,inplace=True)
        self.data.drop(labels=flagged,axis=1,inplace=True)


    def computeFoldChange(self,subtract_baseline=True):
        '''
        Computes the fold change for all wells using the object's unmodified raw data. The object's key
            must have the following columns ['Plate_ID','Gropu','Control']. Control values must be {0,1}.
            The fold change is computed using measurements that have had the first measurment (first time-
            point subtracted, first. The maximum measurement in controls are averaged to get the 447r 
            (i.e. the average maximum OD of control wells) which divides the maximum OD of all cases.
            Fold-changes are normalized to controls belonging to the same group, all wells in a Biolog plate
            will belong to the same group and have the same control (A1 well).
        '''

        mapping = self.key.copy()


        # if mapping lacks group and control columns skip
        if ('Group' not in mapping.keys()) or ('Control' not in mapping.keys()):

            mapping.loc[:,'Fold_Change'] = [np.nan]*mapping.shape[0]

            self.key = mapping

            return None         

        df = self.input_data.copy()  # timepoints by wells, input data that remains unmodified

        # subtract first time point from each column (i.e. wells) 
        if subtract_baseline:
            baseline = df.iloc[0,:]
            df = df.apply(lambda row: row - baseline,axis=1)

        # find all unique groups
        plate_groups = mapping.loc[:,['Plate_ID','Group']].drop_duplicates()
        plate_groups = [tuple(x) for x in plate_groups.values]

        for plate_group in plate_groups:
            
            pid,group = plate_group

            # grab lists of Sample_ID of wells corresponding to control and cases
            controls = subsetDf(mapping,{'Plate_ID':[pid],'Group':[group],'Control':[1]}).index.values
            cases = subsetDf(mapping,{'Plate_ID':[pid],'Group':[group],'Control':[0]}).index.values

            # if group does not have a control, skip
            if len(controls)==0:

                mapping.loc[cases,'Fold_Change'] = [np.nan]*len(cases)

                continue

            df_controls = df.loc[:,controls]
            df_cases = df.loc[:,cases]

            # for denominator, max by control column (i.e. well), then average all controls
            df_controls_fc = df_controls.max(axis=0,numeric_only=True) / df_controls.max(axis=0,numeric_only=True).mean(axis=0,numeric_only=True)
            df_cases_fc = df_cases.max(axis=0,numeric_only=True) / df_controls.max(axis=0,numeric_only=True).mean(axis=0,numeric_only=True)

            mapping.loc[controls,'Fold_Change'] = df_controls_fc
            mapping.loc[cases,'Fold_Change'] = df_cases_fc

        self.key = mapping


    def convertTimeUnits(self,input,output):
        '''
        Converts object's time pandas.DataFrame values between different time intervals. 
            It can convert between seconds, minutes, hours, and days.

        Args:
            input (str): either 'seconds','minutes','hours','days' (current unit of time)
            output (str): either 'seconds','minutes','hours','days' (desired unit of time)
        '''

        seconds = 1.0
        minutes = 60.0  # seconds
        hours = 3600.0  # seconds
        days  = 3600.0 * 24  # second

        time_dict = {'seconds':seconds,'minutes':minutes,'hours':hours,'days':days}

        scaler = time_dict[input] / time_dict[output]
        self.time = self.time.astype(float) * scaler


    def raiseData(self):
        '''
        Replace non-positive (i.e. negative or zero) values in each sample measurement with
            lowest positive meausrement in sample. This is necessary if data will be later
            log-transformed otherwise errors will occur.
        '''

        # raise any negative or zero values to a pseudo-count or baseline OD. 
        self.data = self.data.apply(lambda x: handle_non_pos(x))
        self.mods.raised = True

        # add OD_Offset to key
        self.key.loc[:,'OD_Offset'] = self.data.iloc[0,:].values


    def logData(self,to_do=True):
        '''
        Transform with a natural logarithm all values in object's data (pandas.DataFrame).
        '''

        if not to_do:
            return None

        self.data = self.data.apply(lambda x: np.log(x))
        self.mods.logged = True


    def subtractBaseline(self,to_do=True,poly=False,groupby=None):
        '''
        Subtract the first value in each column from all values of the column.

        WARNING: if performed after natural-log-transformation of data, this is equivalent to 
             scaling relative to OD at T0 (i.e. log[OD(t)] - log[OD(0)] = log[ OD(t) / OD(0) ] )
        '''

        if not to_do:
            return None

        if poly:
            p,ind = (3,5) #(polynomial degrees, num time points to use)
            if groupby is None:
                groups = {None:tuple(self.key.index)}
            else:
                groups = self.key.groupby(groupby).groups

            time = self.time.iloc[:ind].values.ravel()

            for k,index in groups.items():
                temp = self.data.loc[:,index]
                od = temp.values[:ind,:]
                coeff = np.polyfit(time*od.shape[1],od,p)
                temp = temp - np.polyval(coeff,self.time.values[0])
                self.data.loc[:,index] = temp
        else:
            baseline = self.data.iloc[0,:]
            self.data = self.data.apply(lambda x: x - baseline,axis=1)

        self.mods.floored = True


    def isSingleMultiWellPlate(self):
        '''
        Checks if object describes a 96-well plate using several criteria. 

        Returns:
            (boolean)
        '''

        # must have only one Plate_ID associated with all samples
        if len(self.key.Plate_ID.unique()) != 1:
            return False	

        # Well must be a column, values would be well locations (e.g. A1)
        if 'Well' not in self.key.columns:
            return False

        # makes sure that all 96 well locations are described in key and thus data
        expc_wells = set(parseWellLayout(order_axis=1).index.values)  # expected list of wells 
        list_wells = set(self.key.Well.values)  # actual list of wells

        if len(list_wells) != 96:
            return False

        if len(expc_wells.intersection(list_wells)) == 96:
            return True


    def saveKey(self,save_path):
        '''
        Saves the current instance of the object's key (i.e. mapping file). 

        Args:
            save_path (str): file path to location where key should be stored, includes file name. 

        Actions:
            Saves a tab-separated file in desired location (argument).
        x'''

        self.key.to_csv(save_path,sep='\t',header=True,index=True)


    def plot(self,save_path='',plot_fit=False,plot_derivative=False,plot_raw_with_fit=False):
        '''
        Creates a 8x12 grid plot (for 96-well plate) that shows the growth curves in each well.
            Plot aesthetics require several parameters that are saved in config.py and pulled using 
            functions in misc.py. Plot will be saved as a PDF to location passed via argument. Index
            column for object's key should be Well IDs but object's key should also have a Well column.

        Args:
            save_path (str): file path: if empty, plot will not be saved at all.
            plot_fit (boolean): whether to plot GP fits on top of raw OD.
            plot_derivative (boolean): if True, plot only the derivative of GP fit instead. 

        Returns:
            fig,axes: figure and axis handles.

        Action:
            if user passes save_path argument, plot will be saved as PDF in desired location 
        '''

        sns.set_style('whitegrid')
        fontsize=15

        time = self.time

        # make sure plate is 96-well version, otherwise skip plotting
        if not self.isSingleMultiWellPlate():
            msg = f'WARNING: GrowthPlate() object for {self.key.Plate_ID.iloc[0]} is not a 96-well plate. '
            msg += 'AMiGA can not plot it.\n'
            print(msg)
            return None

        self.addLocation()

        #key = self.key
        cols =['Sample_ID','Plate_ID','Well','Row','Column','Fold_Change','OD_Max','OD_Baseline','Flag']
        key = self.key.reindex(cols,axis='columns',)
        key = key.dropna(axis=1,how='all')
        if 'Sample_ID' in key.columns:
            key = key.drop_duplicates().set_index('Sample_ID')
        
        if plot_derivative: 
            base_y = self.gp_data.pivot(columns='Sample_ID',index='Time',values='GP_Derivative')
        elif plot_fit:
            base_y = self.gp_data.pivot(columns='Sample_ID',index='Time',values='OD_Growth_Data')
            overlay_y = self.gp_data.pivot(columns='Sample_ID',index='Time',values='OD_Growth_Fit')
        elif plot_raw_with_fit:
            base_y = self.gp_data.pivot(columns='Sample_ID',index='Time',values='OD_Data')
            overlay_y = self.gp_data.pivot(columns='Sample_ID',index='Time',values='OD_Fit')
        else:
            base_y = self.data#gp_data.pivot(columns='Sample_ID',index='Time',values='OD_Data')

        fig,axes = plt.subplots(8,12,figsize=[12,8])

        # define window axis limits
        ymax = np.ceil(base_y.max(axis=1,numeric_only=True).max(numeric_only=True))
        if plot_derivative:
            ymin = np.floor(base_y.min(axis=1,numeric_only=True).min(numeric_only=True))
        else:
            ymin = 0

        if plot_fit:
            ymin = 0

        xmin = 0
        xmax = np.ravel(time)[-1]
        xmax_up = int(np.ceil(xmax)) # round up to nearest integer

        for well in base_y.columns:

            # select proper sub-plot
            r,c = key.loc[well,['Row','Column']] -1
            ax = axes[r,c]

            # highlight flagged well
            if key.loc[well,'Flag'] and getValue('plot_flag_wells')=='empty':

                [ax.spines[ii].set(lw=0) for ii in ['top','bottom','right','left']]
                plt.setp(ax,xticks=[],yticks=[])

                continue

            # get colors based on fold-change and uration parameters
            if 'Fold_Change' in key.keys():
                color_l,color_f = getPlotColors(key.loc[well,'Fold_Change'])
            else:
                color_l = getValue('fcn_line_color')
                color_f = getValue('fcn_face_color')

            # set window axis limits
            ax.set_xlim([xmin,xmax])
            ax.set_ylim([ymin,ymax])

            # define x-data and y-data points
            x = np.ravel(time.values)
            y = base_y.loc[:,well].values

            # plot line and fill_betwen, if plotting OD estimate
            ax.plot(x,y,color=color_l,lw=1.5,zorder=10)
            if not plot_derivative:
                ax.fill_between(x=x,y1=[ax.get_ylim()[0]]*len(y),y2=y,color=color_f,zorder=7)

            # add fit lines, if desired
            if plot_fit or plot_raw_with_fit:
                y_fit = overlay_y.loc[:,well].values
                ax.plot(x,y_fit,color=getValue('gp_line_fit'),alpha=0.65,ls='--',lw=1.5,zorder=10)

            # show tick labels for bottom left subplot only, so by default no labels
            if plot_derivative:
                plt.setp(ax,yticks=[ymin,0,ymax],yticklabels=[])  # zero derivative indicates no instantaneous growth
            else:
                plt.setp(ax,yticks=[ymin,ymax],yticklabels=[])
            plt.setp(ax,xticks=[xmin,xmax],xticklabels=[])

            # add well identifier on top left of each sub-plot
            well_color = getTextColors('Well_ID')
            ax.text(0.,1.,key.loc[well,'Well'],fontsize=10,color=well_color,ha='left',va='top',transform=ax.transAxes)

            # add Max OD (or Max dOD/dt) value on top right of each sub-plot
            if plot_derivative:
                od_max = np.max(y)
            elif self.mods.floored:
                od_max = key.loc[well,'OD_Max'] - key.loc[well,'OD_Baseline']
            else:
                od_max = key.loc[well,'OD_Max']
            ax.text(1.,1.,f"{od_max:.2f}",fontsize=10,
                color=getTextColors('OD_Max'),ha='right',va='top',transform=ax.transAxes)
            
            if key.loc[well,'Flag'] and getValue('plot_flag_wells')=='cross':
            
                kwargs = {'color':'red','lw':2,'ls':'-','zorder':5}
                (xminf,xmaxf), (yminf,ymaxf) = ax.get_xlim(), ax.get_ylim()
                ax.plot([xminf,xmaxf],[yminf,ymaxf],**kwargs)
                ax.plot([xminf,xmaxf],[ymaxf,yminf],**kwargs)
                plt.setp(ax,xticks=[],yticks=[])


        # show tick labels for bottom left sub-plot only
        plt.setp(axes[7,0],xticks=[0,xmax],xticklabels=[0,xmax_up])
        plt.setp(axes[7,0],yticks=[ymin,ymax],yticklabels=[ymin,ymax])
        largeTickLabels(axes[7,0],fontsize=fontsize)
        
        # add x- and y-labels and title
        ylabel_base = getValue('grid_plot_y_label')
        #ylabel_mod = ['ln ' if self.mods.logged else ''][0]
        ylabel_mod = ''

        if plot_derivative:
            ylabel_text = f'd[ln{ylabel_base}]/dt'
        else:
            ylabel_text = ylabel_mod + ylabel_base

        # add labels and title 
        fig.text(0.512,0.07,'Time ({})'.format(getTimeUnits('output')),fontsize=fontsize,ha='center',va='bottom')
        fig.text(0.100,0.50,ylabel_text,fontsize=fontsize,ha='right',va='center',rotation='vertical')
        fig.suptitle(x=0.512,y=0.93,t=key.loc[well,'Plate_ID'],fontsize=fontsize,ha='center',va='center')

        # if no file path passed, do not save 
        if save_path!='':
            plt.savefig(save_path, bbox_inches='tight')

        self.key.drop(labels=['Row','Column'],axis=1,inplace=True)

        plt.close()

        return fig,axes


    def addLocation(self):
        '''
        Expands object key to include following columns: Row and Column, which describe
            the location of each sample in a multi-well plate (e.g. 96-well plate). However,
            locations here are numerical (i.e Rows 1-8 will correspond to A-H, respectively. 

        Action:
            self.key will have two additional columns, if they were missing prior to execution.
        '''

        cond_1 = all(x in self.key.columns for x in ['Row','Column'])
        if cond_1:
            cond_2 = all([isinstance(ii,int) for ii in self.key.Column.values])
            cond_3 = all([isinstance(ii,int) for ii in self.key.Row.values])
            if cond_2 and cond_3: 
                return None
            else:
                self.key = self.key.drop(labels=['Row','Column'],axis=1)

        if 'Well' in self.key.columns:
            self.key = self.key.join(parseWellLayout(order_axis=1),on='Well')
        else:
            self.key = self.key.join(parseWellLayout(order_axis=1).reset_index())

        row_map = {'A':1,'B':2,'C':3,'D':4,'E':5,'F':6,'G':7,'H':8}
        col_map = {ii:int(ii) for ii in self.key.Column.values}

        self.key.Row = self.key.Row.replace(row_map).infer_objects(copy=False)
        self.key.Column = self.key.Column.replace(col_map).infer_objects(copy=False)


    def extractGrowthData(self,args_dict={},unmodified=False):
        '''
        Creates a GrowthData() object that is equal or smaller based on user-passed criteria.

        Args:
            args_dict (dictionary): keys are experimental variables and values are instances of these variables
            unmodified (boolean): whether to extract data and only pass the input unmodified
                data and time attributes from parent. 

        Returns:
            obj (GrowthPlate object): see definition of parent class for more details.
        '''

        # make sure criteria for selecting data based on experimental variables is not empty
        if not bool(args_dict):
            msg = "USER ERROR: To extract selected growth data, "
            msg += "you must pass criteria for selection to GrwothPlate().extractGrowthData()."
            print(msg)
            return None

        # make sure that criteria for selecting data is formatted as a dictionary of lists or np.arrays
        for dict_key,dict_value in args_dict.items():
            if isinstance(dict_value,(list,np.ndarray)):
                continue
            elif isinstance(dict_value,(int,float,str)):
                args_dict[dict_key] = [dict_value]

        sub_key = self.key
        sub_key = sub_key[sub_key.isin(args_dict).sum(axis=1,numeric_only=True)==len(args_dict)]
        sub_mods = self.mods

        sub_input_data = self.input_data.loc[:,sub_key.index]
        sub_input_time = self.input_time

        if unmodified:
            sub_data = sub_input_data
            sub_time = sub_input_time
        else:
            sub_data = self.data.loc[:,sub_key.index]
            sub_time = self.time

        # if any row in data contains at least one np.nan value, it will be dropped
        sub_data = sub_data.dropna(axis=0,how='any')
        sub_time = sub_time.loc[sub_data.index,:]

        # package as a GrowthPlate object
        obj = GrowthPlate(sub_data,sub_key,sub_time,sub_input_time,sub_input_data,sub_mods)

        return obj


    def copy(self):
        '''
        Creates a copy of class instance.

        Returns:
            obj (GrowthPlate object): see definition of parent class for more details.
        '''

        return deepcopy(self)


    def compute_k_error(self):
        '''Compute K error for checking quality of fit'''

        def foo(x,thresh):

            if x['expected'] == 0 and x['predicted'] == 0:
                kerr =  0
            elif x['expected'] == 0 and x['predicted'] != 0:
                kerr = np.inf
            else:
                kerr = abs((x['predicted']/x['expected'])-1) * 100

            if kerr > thresh:
                return 'TRUE'
            else:
                return 'FALSE'
        
        # whetehr to use use raw or adjusted OD
        if 'Adj_OD_Max' in self.key.columns:
            refs = ['Adj_OD_Max','Adj_OD_Baseline']
        else:
            refs = ['OD_Max','OD_Baseline']
        thresh = getValue('k_error_threshold')

        # compute K_Error and flag ones that deviate above threshold
        sub_df = pd.DataFrame(index=self.key.index,columns=['expected','predicted'])
        sub_df.loc[:,'expected'] = (self.key.loc[:,refs[0]] - self.key.loc[:,refs[1]]).values
        if self.mods.logged:
            sub_df.loc[:,'predicted'] = self.key.k_lin.values
        else:
            sub_df.loc[:,'predicted'] = self.key.k_log.values
        sub_df.loc[:,'K_Error'] = sub_df.apply(lambda x: foo(x,thresh),axis=1)
        
        self.key.loc[:,f'K_Error > {thresh}%'] = sub_df.loc[:,'K_Error']


    def model(self,nthin=1,store=False,verbose=False):
        '''
        Infers growth parameters of interest (including diauxic shifts) by Gaussian Process fitting of data.

        Args:
            store (boolean): if True, certain data will be store as object's attributes
            diauxie (float): ratio of peak height (relative to maximum) used to call if diauxie occured or not

        Actions:
            modifies self.key, and may create self.latent and self.dlatent_dt objects
        '''

        # initialize variables for storing parameters and data
        data_ls, diauxie_dict = [], {}
        gp_params = initParamDf(self.key.index,0)
        mse_df = pd.DataFrame(index=self.key.index,columns=['MSE'])

        for sample_id in self.key.index:

            pid,well = self.key.loc[sample_id,['Plate_ID','Well']].values

            smartPrint(f'Fitting {pid}\t{well}',verbose)

            # extract sample
            args_dict = self.key.loc[sample_id,['Well','Plate_ID']].to_dict()
            sample = self.extractGrowthData(args_dict)

            df = sample.time.join(sample.data)
            df.columns = ['Time','OD']

            # create GP object and analyze
            gm = GrowthModel(df=df,
                             baseline=sample.key.OD_Offset.values,
                             ARD=False,heteroscedastic=False,nthin=nthin,logged=self.mods.logged)

            curve = gm.run(name=sample_id)

            mse_df.loc[sample_id,:] = curve.compute_mse()
            diauxie_dict[sample_id] = curve.params.pop('df_dx')
            gp_params.loc[sample_id,:] = curve.params

            # passively save data, manipulation occurs below (input OD, GP fit, & GP derivative)
            if store:
                data_ls.append(curve.data())

        smartPrint('',verbose)

        diauxie_df = mergeDiauxieDfs(diauxie_dict)

        # record results in object's key
        self.key = self.key.join(gp_params)
        self.key = self.key.join(mse_df)
        self.key = pd.merge(self.key,diauxie_df,on='Sample_ID')

        # check quality of fit with K_Error
        self.compute_k_error()

        # plotting needs transformed (or real) OD & GP fit, & may need GP derivative, save all as obejct attributes
        if store:
            self.gp_data = pd.concat(data_ls).reset_index(drop=True)

        return None




#!/usr/bin/env python

'''
AMiGA library for diauxic shift detection and characterization.
'''

__author__ = "Firas Said Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (3 functions)

# detectDiauxie
# pad
# mergePhases

import numpy as np # type: ignore
import pandas as pd # type: ignore

from .utils import getValue


def detectDiauxie(x,y0,y1,y2,cov0,cov1,thresh,varb='K'):
    '''
    Decompose a growth curve into individual growth phases separated by OD inflection.
    
    Args:
        x (numpy.ndarray): time
        y0 (numpy.ndarray): mean of latent function
        y1 (numpy.ndarray): mean of first derivative of latent function (i.e. growth rate)
        y2 (numpy.ndarray): mean of second dderivative of latent function (e.g. acceleration)
        cov0 (numpy.ndarray): covariance of latent function
        cov1 (numpy.ndarray): covariance of first derivatie of latent function
        thresh (float): ?
        varb (str): use either 'K' or 'r' to threshold/call secondary growth curves
    
    Retuns:
        ret (pandas.DataFrame): dataframe summarizes each growth phase with following:
            t_left: time at left bound
            t_right: time at right bound
            K: total growth
            r: maximum growth rate
            r_left: growth rate at left bound
            r_right: growth rate at ight bound
    '''

    if varb == 'K':
        second_varb = 'r'
    else:
        second_varb = 'K'

    if x.ndim>1:
        x = x[:,0].ravel() # assumes time is first dimension

    # unravel these numpy.ndarray arguments (avoids issues in line 97) 
    # <-- this may cause problems if x is multi-dimensional [NEED TO VERIFY] -->
    y0, y1, y2 = (ii.ravel() for ii in [y0, y1, y2])

    # indices for inflections
    #ips = list(np.where(np.diff(np.sign(y2.ravel())))[0])
    ips = list(np.where(np.diff(np.sign(y2)))[0])

    cond_1 = len(ips)==0
    cond_2 = (len(ips)==1 and ips[0]==0)
    cond_3 = (len(ips)==1 and ips[0]==len(y2)-1)
    cond_4 = np.max(y0) < getValue('diauxie_k_min')
    
    if cond_1 or cond_2 or cond_3 or cond_4:
        cols = ['t_left','t_right','K','r','r_left','r_right']
        ret = pd.DataFrame([x[0],x[-1],np.max(x),np.max(y1),y1[0],y1[-1]],index=cols)
        return ret.T
    
    # types of inflections
    its = [np.sign(y2[ii+1]) if ii<(len(y2)-2) else -1*np.sign(y2[ii-1]) for ii in ips]

    # pad edge cases 
    ips,its = pad(ips,its,edge=1,length=len(y2))
    ips,its = pad(ips,its,edge=-1,length=len(y2))
    
    # convert data types
    ips = np.array([int(ii) for ii in ips])
    its = np.array(its)

    # define bounds of each growth stage
    starts = np.where(its==1)[0][:-1]
    stops = starts+2

    # initialize a summary dataframe and populate with bounds
    #ret = np.zeros((int(len(ips)/2),7))
    ret = np.zeros((len(starts),7))
    ret[:,0] = ips[starts]
    ret[:,1] = ips[stops]
    
    # compute several metrics for growth stage (should I use absolute?)
    bounds = [(int(ii[0]+1),int(ii[1]+1)) for ii in ret]
    ret[:,2] = [np.max(y0[left:right]-y0[left]) for left,right in bounds] # Total change in OD
    ret[:,3] = [np.max(y1[left:right]) for left,right in bounds]         # max growth rate, 
    ret[:,[4,5]] = [[y1[left-1],y1[right-1]] for left,right in bounds]     # growth rate at both bounds 
    # define attraction of each growth stage: a growth stage is attrached 
    #   to the adjacent gowth stage with the least difference in terms of 
    #   growth rate at the shared bounds (relative to max growth rate 
    #   within the bounds)  
    ret[:,6] = [-1 if np.abs(row[5]-row[3]) > np.abs(row[4] - row[3]) else 1 for row in ret]

    # annotate datafame and sort in ascending order
    cols = ['t_left','t_right','K','r','r_left','r_right','attraction']
    #cols = ['ind0','ind1','y_delta','max_y1','y1(ind0)','y1(ind1)','attraction']
    ret = pd.DataFrame(ret,columns=cols)

    # how to deal with negative r or K
    #   if at least one value is nonzero positive
    if any(ii>0 for ii in ret[varb].values):
        # starting with the smallest growth stage (smallest total change in OD):
        #    if it's K is smaller than a certain proportion of the max K
        #    merge with attractor, continue until all growth phases meet criteria
        while ret[varb].min() < thresh*ret[varb].max() :
            
            ret = ret.sort_values(['t_left'])
            ret.iloc[0,-1] = 1   # first phase is always attracted forward in time
            ret.iloc[-1,-1] = -1 # last phase is always attracted backward in time
            
            ret = ret.sort_values([varb,second_varb])
            idx = ret.index.values[0]
            att = ret.loc[idx,'attraction']
            att = idx+att
            ret = mergePhases(ret,idx,att,varb=varb)
            
            # should you re-compute attraction?
    else:  
        while ret.shape[0] > 1:  # coalescale all into a single curve
            ret = mergePhases(ret,0,1)

    # re-sort by time and convert array indices to time values
    ret = ret.sort_values(['t_left'])
    ret.iloc[:,0] = ret.iloc[:,0].apply(lambda i: x[int(i)])
    ret.iloc[:,1] = ret.iloc[:,1].apply(lambda i: x[int(i)])
    ret.drop('attraction',axis=1,inplace=True)
    
    return ret


def pad(ips,its,edge=1,length=None):
    '''
    pad() is used only within diauxie(). Diauxie functions best if the growth curve
        begins at time zero with a positive inflection and end at last time point
        with another positive inflection. pad() pads the edges of inflection points 
        (ips in time units) and inflection types (its, whether +ve or -ve) with dummy
        inflection points to assist diauxie in analyzing a gowth curve. 
    '''

    # if negative, pad the right edge of the curve, otherwise pad the left
    if edge==-1:
        ips = ips[::-1]
        its = its[::-1]
        idx = length-1
    else:
        idx = 0

    if (ips[0] == 0) and (its[0] == 1):
        pass
    elif (its[0] == -1):
        ips = [idx] + ips
        its = [1] + its
    elif (its[0] == 1):
        ips = [idx,idx] + ips
        its = [1,-1] + its

    # reset the order of the input before retuning
    if edge==-1:
        ips = ips[::-1]
        its = its[::-1]

    return ips,its

def mergePhases(df,row1_idx,row2_idx,varb='K'):
    '''
    mergePhasese() is used only within diauxie(). Diauxie() iterates through
        growth phases (described in df) and decides whether two growth phases
        should be merged and considered as a single phase. mergePhases()
        goes though each attibute (df column) and identifies the final value,
        it drops the rejected growth phase and updates the accepted growth phase. 
        
    Args:
        df (pandas.DataFrame): dataframe summarizes each growth phase with following:
            t_left: time at left bound
            t_right: time at right bound
            K: total growth
            r: maximum growth rate
            r_left: growth rate at left bound
            r_right: growth rate at ight bound 
        row1_idx: index value for rejected growth phase
        row2_idx: index value for accepted growth phase
    
    Return:
        df (pandas.DataFrame): # of rows should be one less than input datafame # of rows
    '''
    
    tmp = df.loc[[row1_idx,row2_idx],:].sort_index()
    tmp = tmp.sort_values(['t_left'])
    
    values = ([tmp['t_left'].min(numeric_only=True),
               tmp['t_right'].max(numeric_only=True),
               tmp['K'].sum(numeric_only=True),
               tmp['r'].max(numeric_only=True),
               tmp['r_left'].values[0],
               tmp['r_right'].values[1],
               tmp.loc[row2_idx,'attraction']])
    
    df.loc[row2_idx,:] = values
    df.drop(row1_idx,axis=0,inplace=True)
    df = df.sort_values(['t_left']).reset_index(drop=True) 
    
    return df



#!/usr/bin/env python

'''
AMiGA wrapper: getting time at which measurements reach a user-defined threshold.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (4 functions)

# main
# update
# find
# read

import numpy as np # type: ignore
import pandas as pd # type: ignore


def main(args):

	tabs = read(args)
	summ = find(args,tabs)
	summ = update(args,summ)


def update(args,summ):

	summ.to_csv(args.summary,sep='\t',header=True,index=False)

def find(args,tables):

    # define arguments
	col = args.curve_format
	thresh = args.threshold

	data, summ = tables

	ls_time = []
	sample_ids = summ.Sample_ID.unique()

	for sid in sample_ids:

		tmp = data[data.Sample_ID==sid]
		tmp = tmp.loc[:,['Time',col]]

		matches = tmp[tmp[col]>=thresh].Time.values

		if len(matches) == 0:
			ls_time.append(np.inf)
		else:
			ls_time.append(matches[0])

	description = f't(od>={thresh})'
	t_od = pd.DataFrame([sample_ids,ls_time],index=['Sample_ID',description]).T

	summ = pd.merge(summ,t_od,on='Sample_ID')

	return summ


def read(args):

	data = pd.read_csv(args.gp_data,sep='\t',index_col=0)
	summ = pd.read_csv(args.summary,sep='\t')

	tables = (data,summ)

	return tables




import sys
import os

# Add the parent directory to sys.path to allow package imports
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

from amiga.amiga import AMiGA

if __name__ == "__main__":
    AMiGA()


#!/usr/bin/env python

'''
AMiGA library for handling and reporting growth parameters.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (8 functions)

# articulateParameters
# initDiauxieList
# initParamList
# initParamDf
# mergeDiauxieDf
# minimizeParameterReport
# minimizeDiauxieReport
# prettifyParameterReport
# removeFromDiauxieReport
# removeFromParameterReport

import sys
import pandas as pd # type: ignore

from scipy.stats import norm # type: ignore

from .utils import getValue


def initDiauxieList(params=None):
    '''
    Returns list of strings which are growth parameters specific to diauxie characterization.
    '''

    p = initParamList(0,params=params)
    p = [f'dx_{ii}' for ii in p]
    p += ['dx_t0','dx_tf']
    p.remove('dx_diauxie')

    return p


def initParamList(complexity=0,params=None):
    '''
    Initialize a list of parameter variables. 

    Args:
        complexity (0,1,2): whether to include raw, mean and std, or norm version of parameters
        params(None or list): list of parameters to process

    Returns (list)
    '''

    if params is None:
        lp0 = ['auc_lin','auc_log','k_lin','k_log','death_lin','death_log',
               'gr','dr','td','lagC','lagP',
               't_k','t_gr','t_dr','diauxie']
    else:
        lp0 = params

    if isinstance(lp0,str):
        lp0 = [lp0]

    # create variation of each parameter: i.e, mean(), std(), and norm()

    # first get rid of diauxie
    lp0_copy = lp0.copy()
    if 'diauxie' in lp0_copy:
        lp0_copy.remove('diauxie')

    lp1a = [f'mean({ii})' for ii in lp0_copy]
    lp1b = [f'std({ii})' for ii in lp0_copy]
    lp2 = [f'norm({ii})' for ii in lp0_copy]

    if complexity == 0:
        return lp0
    elif complexity == 1:
        return lp1a + lp1b
    elif complexity == 2:
        return lp2


def initParamDf(sample_ids=None,complexity=0):
    '''
    Initialize lists of parameter names or empty pandas.DataFrame to store growth parameers.

    Args:
        sample_ids (list): used to index dataFrame rows.
        complexity (0,1,2): whether to include raw, mean and std, or norm version of parameters

    Return:
        (pandas.DataFame,pandas.DataFrame)
    '''

    params = initParamList(complexity=complexity)

    for ii in ['mean(diauxie)','std(diauxie)']:
        if ii in params:
            params.remove(ii)

    return pd.DataFrame(index=sample_ids,columns=params)


def mergeDiauxieDfs(diauxie_ls):
    '''
    Merges list of pandas.DataFrames into a single one vertically.

    Args (list)
    Returns (pandas.DataFrame)
    '''

    def addSampleID(df,sid):
        '''
        Adds Sample ID (str) as a column in a pandas.dataframe and returns the dataframe.
        '''
        df.index = [sid]*df.shape[0]
        df.index.name = 'Sample_ID'
        df = df.reset_index(drop=False)
        return df

    merged = {sid:addSampleID(df,sid) for sid,df in diauxie_ls.items()}
    merged = pd.concat(merged.values(),axis=0).reset_index(drop=True)

    return merged

def minimizeParameterReport(df):
    '''
    Minimizes a pandas.DataFrame to only inlcude parameters indicated in 
        the config.py file under 'report-parameters' variable. 

    Args (pandas.DataFrame)
    Return (pandas.DataFrame)
    '''

    request = getValue('report_parameters')
    request = initParamList(0,request) + initParamList(1,request) + initParamList(2,request)

    lp = initParamList(0) + initParamList(1) + initParamList(2)
    keys = set(lp).intersection(set(df.keys()))
    remove = keys.difference(set(request))

    df = df.drop(labels=remove,axis=1)

    return df


def minimizeDiauxieReport(df):
    '''
    Minimizes a pandas.DataFrame to only inlcude parameters indicated in 
        the config.py file under 'report-parameters' variable. 

    Args (pandas.DataFrame)
    Return (pandas.DataFrame)
    '''

    request = getValue('report_parameters')
    request = initDiauxieList(request)

    lp = initDiauxieList()
    keys = set(lp).intersection(set(df.keys()))
    remove = keys.difference(set(request))

    return df.drop(labels=remove,axis=1)


def removeFromParameterReport(df,to_remove=None):
    
    # validate input type
    if to_remove is None: 
        return df
    elif isinstance(to_remove,str): 
        to_remove = [to_remove]
    
    # all variations of parameters to remove
    to_remove = initParamList(0,to_remove) + initParamList(1,to_remove) + initParamList(2,to_remove)
    
    # all parameters that are possibly produced by AMiGA
    lp = initParamList(0) + initParamList(1) + initParamList(2)

    # all parameters that are possible but exists in input dataframe
    keys = set(lp).intersection(df.keys())
    
    # which of those parameters in dataframe must be removed
    to_remove = keys.intersection(set(to_remove))

    return df.drop(labels=to_remove,axis=1)


def removeFromDiauxieReport(df,to_remove=None):
    
    # validate input type
    if to_remove is None: 
        return df
    elif isinstance(to_remove,str): 
        to_remove = [to_remove]
    
    # all variations of parameters to remove
    to_remove = [f'dx_{ii}' for ii in initParamList(0,to_remove)]
    
    # all parameters that are possibly produced by AMiGA
    lp = initDiauxieList()

    # all parameters that are possible but exists in input dataframe
    keys = set(lp).intersection(df.keys())
    
    # which of those parameters in dataframe must be removed
    to_remove = keys.intersection(set(to_remove))

    return df.drop(labels=to_remove,axis=1)


def prettyifyParameterReport(df,target,confidence=0.975):
    '''
    Called by compare.py to create a human-readable comparison of parameter estimates
        between two user-selected conditions. The generated table will have parameters 
        as index column, the first few rows will be meta-data, the cell values will be 
        either the mean or confidence interval of estimates. 

    Args:
        df (pandas.DataFrame): Two samples (rows) by many growth parameters (columns).
            The growth parameters can be summay (mean and standard deviation) estimates 
            or simple estimate (mean only).
        target (string)
        confidence (float): must be between 0.8 and 1.
    '''

    if df.shape[0] != 2:

        msg = 'FATAL USER ERROR: AMiGA can only contrast the confidence intervals of '
        msg += 'two experimental conditions. Please subest the input dataframe properly '
        msg += 'before submitting to prettifyParameterReport.'
        sys.exit(msg)


    def getConfInts(means,stds,z_value=0.975):
        '''
        Computes confidence interval based on mean, standard deviation, and desired confidence.

        Args:
            means (array of floats)
            stds (array of floats)
            z_value (float)

        Returns:
            cis (array of strings), where each formatted string indicates the confidence interval,
                e.g. [0.4,0.6]
        '''

        scaler = norm.ppf(z_value)

        cis = []
        for m,s in zip(means,stds):
            low, upp = m-scaler*s, m+scaler*s
            cis.append(f'[{low:.3f},{upp:.3f}]')

        return cis

    def detSigDiff(a,b):
        '''
        Detemines if there is a significant difference between two variables, 
            based on whether confidence intervals overlap or not

        Args:
            a (2-array): lower and upper bounds of first confidence interval
            b (2-array): lower and upper bounds of second confidence intervals

        Returns:
            (boolean): True (intervals do not overlap) or False (intervals overlap)
        '''

        a = [float(ii) for ii in a]
        b = [float(ii) for ii in b]
        return not ((a[0] <= b[1]) and (b[0] <= a[1]))

    alpha = 1-confidence
    z_value = 1-(alpha/2)

    df = df.set_index(['Sample_ID'])

    params = list({ii.split('(')[1][:-1] if '(' in ii else ii for ii in df.T.index[1:]})

    df_mus = pd.DataFrame(columns=df.iloc[:,0],index=params).sort_index()
    df_cis = pd.DataFrame(columns=list(df.iloc[:,0])+[''],index=params).sort_index()
    df_cis.T.index.name = target

    for p in params:

        if p == 'diauxie':
            df_mus.loc[p,:] = df.loc[:,p].values
            if df.loc[:,p].values[0] != df.loc[:,p].values[1]:
                df_cis.loc[p,:] = ['NA','NA',True]
            else:
                df_cis.loc[p,:] = ['NA','NA',False]
        else:
            mus = df.loc[:,f'mean({p})'].values
            stds = df.loc[:,f'std({p})'].values
            cis = getConfInts(mus,stds,z_value)
            olap = detSigDiff(eval(cis[0]),eval(cis[1]))

            df_mus.loc[p,:] = [f'{ii:.3f}' for ii in mus]
            df_cis.loc[p,:] = cis + [olap]

    df_mus.loc['Parameter',:] = ['Mean','Mean']
    df_cis.loc['Parameter',:] = [f'{100*confidence}% CI',
                                 f'{100*confidence}% CI',
                                 'Sig. Diff.']

    df_mus = df_mus.T.reset_index().T
    df_cis = df_cis.T.reset_index().T
    df_all = df_mus.join(df_cis,lsuffix='L',rsuffix='R')
    df_all = df_all.T.set_index([target,'Parameter']).T

    return df_all

def articulateParameters(df,axis=0):
    '''
    Replace shortnames for growth parameters with clear descriptions.

    Args:
        df (pandas.DataFrame)
        axis (0 or 1): whether parameters are index labels (0) or column names (1)

    Retuns:
        df (pandas.DataFrame)
    '''

    params_labels = {'auc_lin' : 'AUC (lin)',
                     'auc_log' : 'AUC (log)',
                     'death_lin' : 'Death (lin)',
                     'death_log' : 'Death (log)',
                     'diauxie' : 'Diauxie',
                     'dr' : 'Death Rate',
                     'gr' : 'Growth Rate',
                     'k_lin' : 'Carrying Capacity (lin)',
                     'k_log' : 'Carrying Capacity (log)',
                     'lagC' : 'Lag Time',
                     'lagP' : 'Adaptation Time',
                     'td' : 'Doubling Time',
                     't_dr' : 'Time at Max. Death Rate',
                     't_gr' : 'Time at Max. Growth Rate',
                     't_k' : 'Time at Carrying Capacity'}

    return df.rename(params_labels,axis=axis)





#!/usr/bin/env python

'''
AMiGA library of functions for object-oriented testing of hypotheses with GP regression.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"

import os
import sys
import operator
import numpy as np # type: ignore
import pandas as pd # type: ignore
import matplotlib.pyplot as plt # type: ignore
import seaborn as sns # type: ignore

from scipy.stats import norm, percentileofscore # type: ignore
from matplotlib import rcParams # type: ignore
from matplotlib.ticker import MultipleLocator # type: ignore
from tabulate import tabulate # type: ignore

from .detail import updateMappingControls, shouldYouSubtractControl
from .model import GrowthModel
from .growth import GrowthPlate
from .org import assembleFullName, assemblePath
from .trim import annotateMappings,trimMergeMapping, trimMergeData
from .utils import subsetDf, reverseDict
from .utils import getValue, getTimeUnits, getHypoPlotParams, selectFileName
from .comm import prettyNumberDisplay, smartPrint, tidyDictPrint, tidyMessage
from .params import initDiauxieList, initParamList, initParamDf, mergeDiauxieDfs
from .params import minimizeParameterReport, minimizeDiauxieReport
from .params import removeFromParameterReport, removeFromDiauxieReport
from .params import articulateParameters, prettyifyParameterReport
from .plot import addRealPlotLine, addMVNPlotLine, setAxesLabels
from .plot import dynamicWindowAdjustment, plotDeltaOD, savePlotWithLegends

pd.set_option('future.no_silent_downcasting', True)


class HypothesisTest:

    def __init__(self,mapping_dict=None,data_dict=None,params_dict=None,args_dict=None,directory_dict=None,sys_exit=True):
        '''
        Framework for hypothesis testing. 

        Args:
            mapping_dict (dictionary of mapping as pandas.DataFrames)
            data_dict (dictionary of data as pandas.DataFrames)
            params_dict (dictionary)
            args_dict (dictionary)
            directory_dict (dictionary)
            sys_exit (boolean)
        '''

        # define attibutes
        self.args = args_dict
        self.params = params_dict # need keys 'hypothesis' and 'subset' 
        self.directory = directory_dict
        self.hypothesis = params_dict['hypothesis']
        self.subtract_control = self.args.subtract_control
        self.subtract_blanks = self.args.subtract_blanks
        self.verbose = self.args.verbose

        # only proceed if hypothesis is valid
        if self.checkHypothesis():
            return None

        # undertake hypothesis test
        self.initPaths()
        self.defineMapping(mapping_dict=mapping_dict)
        self.addInteractionTerm()
        self.checkControlSamples(mapping_dict=mapping_dict)
        self.defineData(data_dict=data_dict)
        self.prettyTabulateSamples()
        self.prepRegressionPlate()
        self.tidifyRegressionData()
        self.factorizeCategoicals()
        self.executeRegression()
        self.generatePredictions()
        self.savePredictions()
        self.computeFullDifference()
        self.plotPredictions()
        self.reportRegression()
        self.exportReport()

        # bid user farewell
        if sys_exit:
            sys.exit(tidyMessage('AMiGA completed your request!'))


    def initPaths(self):
        '''
        Initialize paths for for saving data and results. 
        '''

        # if user did not pass file name for output, use time stamp
        file_name = selectFileName(self.args.output)
        dir_path = assemblePath(self.directory['models'],file_name,'')
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)      

        # running model on transformed results and recording results
        file_path_key = assembleFullName(dir_path,'',file_name,'key','.txt')
        file_path_input = assembleFullName(dir_path,'',file_name,'input','.txt')

        paths_dict={}

        paths_dict['filename'] = file_name
        paths_dict['dir'] = dir_path
        paths_dict['key'] = file_path_key
        paths_dict['input'] = file_path_input

        self.paths_dict = paths_dict


    def checkHypothesis(self):
        '''
        Verifies that a user provided a hypothesis ant that is meets the following crieteria. The alternative
            hypothesis must have only two variables, one being time.

        Args:
            hypothesis (dictionary): e.g. {'H0':['Time'],'H1':['Time','Substrate']}

        Returns
            hypotheis (dict): input argument but with 'Time' added to values, if initially missing
            variables (list): variables that differ between the null and alternative hypotheses
        '''

        hypothesis = self.hypothesis.copy()

        if len(hypothesis)==0:
            msg = 'USER ERROR: No hypothesis has been passed to AMiGA via either command-line or text file.\n'
            return True #exit

        # what is the variable of interest based on hypothesis?
        target = list(set(hypothesis['H1'])^(set(hypothesis['H0'])))

        if len(target) > 1:
            msg = 'FATAL USER ERROR: Your null and alternaive hypotheses differ by more than one variables. '
            msg += 'AMiGA can not test for differential growth between conditions that vary by more than one variable.\n'
            sys.exit(msg)
            #msg += 'All of these distinct variables will be permuted for credible interval testing.\n'
            #smartPrint(msg,True)

        # make sure that Time is included in both hypotheses
        if ('Time' not in hypothesis['H0']):
            hypothesis['H0'] = hypothesis['H0'] + ['Time']
            msg = 'USER WARNING: Time was missing from the null hypothesis (H0) but it will be included as a variable.\n'
            smartPrint(msg,True)

        if ('Time' not in hypothesis['H1']):
            hypothesis['H1'] = hypothesis['H1'] + ['Time']
            msg = 'USER WARNING: Time was missing from the alternative hypothesis (H1) but it will be included as a variable.\n'
            smartPrint(msg,True)

        # get all unique variables in hypothesis into a list
        variables = {i for value in hypothesis.values() for i in value}
        variables = variables.difference({'Time'})  # remove Time
        self.non_time_varbs = variables
        variables = ['Time'] + list(variables)  # add back Time to beginning of list, not necessary but helpful
        
        self.variables = variables
        self.target = target
        #self.hypothesis = hypothesis

        return False


    def defineMapping(self,mapping_dict=None):
        '''
        Trim data based on user-passed or default 'subest' and 'flag' parameters
        '''

        # annotate Subset and Flag columns in mapping files, then trim and merge into single dataframe
        mapping_dict = annotateMappings(mapping_dict,self.params,self.verbose)

        self.master_mapping = trimMergeMapping(mapping_dict,self.verbose) # named index: Sample_ID


    def addInteractionTerm(self):
        '''
        If user passed hypothesis with an interaction term (identified by an astersisk), then
            create 
        '''

        # add interaction term, if needed
        mapping = self.master_mapping

        for variable in self.target:

            if ('*' in variable):
                pairs = variable.split("*")
                var_dict = {}
                if ('(' in variable) and (')' in variable):

                    for pair in pairs:
                        var,ctrl = pair.split('(')
                        var_dict[var]=ctrl[:-1]

                    intx = subsetDf(mapping,var_dict).index.values
                    mapping.loc[:,variable] = [0]*mapping.shape[0]
                    mapping.loc[intx,variable] = [1]*len(intx)

                else:

                    df = mapping.loc[:,pairs].drop_duplicates()
                    df.loc[:,variable] = df.apply(lambda x: f'{x[pairs[0]]} x {x[pairs[1]]}',axis=1) 

                    mapping = pd.merge(mapping.reset_index(),df,on=pairs,how='left')
                    mapping = mapping.set_index('Sample_ID')

        self.master_mapping = mapping


    def checkControlSamples(self,mapping_dict=None):

        mm = self.master_mapping
        sc = self.subtract_control 

        if sc:
            sc = shouldYouSubtractControl(mm,self.target)
        
        mm = updateMappingControls(mm,mapping_dict,to_do=sc).dropna(axis=1)
        
        self.master_mapping = mm
        self.subtract_control = sc


    def defineData(self,data_dict=None):

        # grab all data
        self.master_data = trimMergeData(data_dict,self.master_mapping,self.args.skip_first_n,self.verbose) # unnamed index: row number


    def prettyTabulateSamples(self):

        if self.verbose:
            print_map = self.master_mapping.copy().drop(['Subset'],axis=1)
            tab = tabulate(print_map,headers='keys',tablefmt='psql')
            msg = 'The following samples will be used in hypothesis testing:'
            msg = f'\n{msg}\n{tab}\n'
            smartPrint(msg,self.verbose)
            smartPrint(tidyDictPrint({'Hypothesis is':self.hypothesis}),self.verbose)


    def prepRegressionPlate(self):
        '''
        Packages data into a growth.GrowthPlate() object and performs a select number of class functions.

        Args:
            data (pandas.DataFrame): t (number of measurements) by n+1 (number of samples + one column for time)
            mapping (pandas.DataFrame): n (number of samples) by p (number of variables)
            subtract_control (boolean)
            thinning_step (int): how many time points to skip between selected time points. 
        '''

        plate = GrowthPlate(self.master_data,self.master_mapping)
        plate.convertTimeUnits(input=getTimeUnits('input'),output=getTimeUnits('output'))
        plate.subtractControl(to_do=self.subtract_blanks,drop=True,blank=True)
        plate.subtractControl(to_do=self.subtract_control,drop=True,blank=False)
        plate.raiseData()  # replace non-positive values, necessary prior to log-transformation
        plate.logData(to_do=self.args.log_transform)
        plate.subtractBaseline(to_do=True,poly=getValue('PolyFit'),groupby=list(self.non_time_varbs))
        plate.dropFlaggedWells(to_do=True)
        plate.key.to_csv(self.paths_dict['key'],sep='\t',header=True,index=True)  # save model results
        self.plate = plate
        self.ntimepoints = plate.time.shape[0]


    def tidifyRegressionData(self):
        '''
        Prepares a single dataframe for running GP regression.

        Args:
            plate (growth.GrowthPlate)
            hypothesis (dictionary): has two keys 'H0' and 'H1', values are lists of variables
                which are headers of plate.ke pandas.DataFrame
            mthin (int)
            save_path (str): default is None

        Returns:
            data (pandas.DataFrame): long-format
            factor_dict (dictionary): maps unique values for variables to numerical integers
        '''

        save_path = self.paths_dict['input']

        variables = self.variables
        non_times = self.non_time_varbs

        # melt data frame so that each row is a single time measurement
        #   columns include at least 'Sample_ID' (i.e. specific well in a specific plate) and
        #   'Time' and 'OD'. Additioncal column can be explicilty called by user using hypothesis.

        data = (self.plate.time).join(self.plate.data)
        data = pd.melt(data,id_vars='Time',var_name='Sample_ID',value_name='OD')
        data = data.merge(self.plate.key,on='Sample_ID')

        # Handle missing data by dropping them (ie, replicates missing select time points)
        missing = np.unique(np.where(data.iloc[:,1:].isna())[0])
        missing = data.iloc[missing,].Time.values

        tmp = data.sort_values(['Time']).loc[:,['Time','OD']]
        drop_idx = tmp[tmp.isna().any(axis=1)].index
        data = data.drop(labels=drop_idx,axis=0)
        data = data.drop(labels=['Subset'],axis=1)

        if save_path:
            data.to_csv(save_path,sep='\t',header=True,index=True)

        # reduce dimensionality
        data = data.reindex(['OD']+variables,axis=1)
        data = data.sort_values('Time').reset_index(drop=True) # I don't think that sorting matters, but why not
        self.data = data

        value_counts = data.loc[:,list(non_times)].apply(lambda x: len(np.unique(x)),axis=0)
        if any(value_counts >2):
            value_counts = pd.DataFrame(value_counts).reset_index()
            value_counts.columns = ['Variable','# Unique Values']
            tab = tabulate(value_counts,headers='keys',tablefmt='psql')

            msg = '\n USER WARNING: Hypothesis testing should only be performed on binary conditions.'
            msg = f'{msg} See below.\n\n{tab}\n'
            sys.exit(msg)


    def factorizeCategoicals(self):

        data = self.data

        # factorize variable values and store mapping
        self.factor_dict = {}
        for varb in self.variables:
            if varb == 'Time':
                continue
            values = data[varb].unique()
            values_code = range(len(values))
            values_maps = {v:c for v,c in zip(values,values_code)}
            data.loc[:,varb] = data.loc[:,varb].replace(values_maps).infer_objects(copy=False)
            self.factor_dict[varb] = values_maps


    def executeRegression(self):
        '''
        Computes the log Bayes Factor and its null distribution (based on permutation tests).

        Args:
            data (pandas.DataFrame): each row is a single measurement (i.e. time point in a well), columns are variables
                and must include 'Time', 'OD'.
            hypothesis (dictionary): keys must be 'H0' and 'H1', values are lists of variables (must match data keys)
            nperm (int): number ofxec permutations to generate null distribution

        Returns:
            log_BF (float): log Bayes Factor = log (P(H1|D)/P(H0|D))
            null_distribution (list of floats): the null distribution for log Bayes Factor where variable of interest
                was permuted for a certain number of times (based on nperm).
        '''

        verbose = self.verbose
        hypothesis = self.hypothesis
        fix_noise = self.args.fix_noise
        nperm = self.args.number_permutations
        nthin = self.args.time_step_size

        data = self.data

        data0 = data.loc[:,['OD']+hypothesis['H0']]
        data1 = data.loc[:,['OD']+hypothesis['H1']]

        gm0 = GrowthModel(df=data0,ARD=True,heteroscedastic=fix_noise,nthin=nthin,logged=self.plate.mods.logged)
        gm1 = GrowthModel(df=data1,ARD=True,heteroscedastic=fix_noise,nthin=nthin,logged=self.plate.mods.logged)

        gm0,LL0 = gm0.run(predict=False)
        gm1,LL1 = gm1.run(predict=False)
        log_BF = LL1-LL0

        self.log_BF = log_BF
        self.model = gm1
        self.LL0 = LL0
        self.LL1 = LL1
        self.log_BF_null_dist = None

        null_distribution = []
        to_permute = list(set(hypothesis['H1']).difference(set(hypothesis['H0'])))[0]
        for rep in range(nperm):
            smartPrint(f'Permutation #{rep}',verbose)
            null_distribution.append(gm1.permute(to_permute)-LL0)
        smartPrint('',verbose)
        if null_distribution:
            self.log_BF_null_dist = null_distribution


    def generatePredictions(self):
        '''
        A Gaussian Process is modelled for Nx1 response (Y) variable and NxD explantory
            (X) variables, with multiple replicates for each unique set of D variables. 
            generateTestMatrix() identifies all uniques instances of X, disregarding the 
            time variable. Time is however reintroduced but as an evenly spaced vector
            based on pred_num_time_points variable in config.py.

        Args:
            model (GPy.models.gp_regression.GPRegression)
            model_input (list of str): variables must be keys for data
            data (pandas.DataFrame): samples by variables (includes Time)

        Returns:
            x (pandas.DataFrame): full design matrix (includes time)
            x_min (pandas.DataFrame): minimal design matrix (excludes time)
        '''

        data = self.data
        model = self.model
        model_input = self.hypothesis['H1']
        fix_noise = self.args.fix_noise

        # first, generates a dataframe where each row is a unique permutations of non-time  variables
        x = data.loc[:,model_input].drop(labels=['Time'],axis=1)
        x = x.drop_duplicates().reset_index(drop=True)
        x_min = x.copy()
        x['merge_key'] = [1]*x.shape[0]

        # second, sample evenly spaced time points based on min and max time points
        x_time = pd.DataFrame(data.Time.unique(),columns=['Time'])
        x_time['merge_key'] = [1]*x_time.shape[0]

        # combine time and variable permutation dataframes, such that each unique time point 
        #    is mapped to all unique permutations of variabels
        x = pd.merge(x_time,x,on='merge_key').drop(labels=['merge_key'],axis=1)

        if fix_noise:
            sigma_noise =np.ravel(model.error_new)+model.noise
        else:
            sigma_noise = np.ravel([model.noise]*x.shape[0])

        x_new = [tuple(ii) if len(ii)>1 else [ii] for ii in model.x_new]
        x = x.set_index(list(x.keys())).loc[x_new].reset_index() # reorder x to match noise

        # given gp model, predict mu and Sigma on experimental design input
        mu,cov = (model.model).predict(x.values,full_cov=True,include_likelihood=False)
        mu_var = pd.DataFrame([np.ravel(mu),np.ravel(np.diag(cov)),np.ravel(sigma_noise)],
                              index=['mu','Sigma','Noise']).T
        x = x.reset_index(drop=True).join(mu_var)

        self.x_full = x
        self.x_min = x_min


    def savePredictions(self):
        '''
        Given model predictions of growth curves (for each unique set of conditions tested),
            describe the latent function and its derivative in terms of growth parameters. 
            Reports results in a file with {file_name}_params name in dir_path directory. 

        Args:
            model (GPy.models.gp_regression.GPRegression)
            data (pandas.DataFrame)
            hypothesis (dictionary): e.g. {'H0':['Time'],'H1':['Time','Substrate']}
            actor_dict (dictionary): mapping of unique values of variables to numerical integers
            posterior (boolean)
            save_latent (boolean)
            dir_path (str): path to directory
            file_name (str): file name

        Returns:
            x_full (pandas.DataFrame): 
            x_min (pandas.DataFrame):

        '''

        model = self.model
        factor_dict = self.factor_dict
        variable = self.target[0]
        confidence = self.args.confidence  # confidence interval, e.g. 0.95

        posterior = self.args.sample_posterior
        save_latent = self.args.save_gp_data
        do_not_log_transform = self.args.do_not_log_transform

        dir_path = self.paths_dict['dir']
        file_name = self.paths_dict['filename']

        # define hypothesis paraameters
        x_full = self.x_full
        x_min = self.x_min

        diauxie_dict = {}
        params_latent = initParamDf(x_min.index,complexity=0)
        params_sample = initParamDf(x_min.index,complexity=1)

        for idx,row in x_min.iterrows():

            # get x and y data
            df = subsetDf(x_full.drop(labels=['mu','Sigma','Noise'],axis=1),row.to_dict())

            # get curve based on model predictions
            gm = GrowthModel(model=model.model,x_new=df.values,ARD=True,logged=model.logged)
            curve = gm.run()

            # get parameter estimates using predicted curve
            diauxie_dict[idx] = curve.params.pop('df_dx')
            params_latent.loc[idx,:] = curve.params

            if posterior:
                params_sample.loc[idx,:] = curve.sample().posterior

        # summarize diauxie results
        diauxie_df = mergeDiauxieDfs(diauxie_dict)

        if posterior:
            gp_params = params_sample.join(params_latent['diauxie'])
        else:
            gp_params = params_latent

        gp_params = x_min.join(gp_params)
        gp_params.index.name = 'Sample_ID'
        gp_params = gp_params.reset_index(drop=False)
        gp_params = pd.merge(gp_params,diauxie_df,on='Sample_ID')

        # save gp_data fit
        x_out = x_full.copy()
        for key,mapping in factor_dict.items():
            if key in x_out.keys():
                mapping = {str(k):str(v) for k,v in mapping.items()}
                x_out.loc[:,key] = x_out.loc[:,key].replace(reverseDict(mapping)).astype(int)
            if key in gp_params.keys():
                mapping = {str(k):str(v) for k,v in mapping.items()}
                gp_params.loc[:,key] = gp_params.loc[:,key].replace(reverseDict(mapping)).astype(int)

        #params = initParamList(0)
        diauxie = initDiauxieList()
        params = initParamList(0) + initParamList(1)
        params = list(set(params).intersection(set(gp_params.keys())))

        df_params = gp_params.drop(labels=diauxie,axis=1).drop_duplicates()
        df_params = minimizeParameterReport(df_params)
        df_diauxie = gp_params[gp_params.diauxie==1].drop(labels=params,axis=1)
        df_diauxie = minimizeDiauxieReport(df_diauxie)

        # because pooling, drop linear AUC, K, and Death 
        to_remove = ['death_lin','k_lin','auc_lin']
        if do_not_log_transform:
            to_remove += ['td']

        df_params = removeFromParameterReport(df_params,to_remove)
        df_diauxie = removeFromDiauxieReport(df_diauxie,to_remove)

        if do_not_log_transform:
            columns = {'auc_log':'auc_lin','k_log':'k_lin','death_log':'death_lin',
                       'dx_auc_log':'dx_auc_lin','dx_k_log':'dx_k_lin','dx_death_log':'dx_death_lin'}
            df_params.rename(columns=columns,inplace=True)
            df_diauxie.rename(columns=columns,inplace=True)

        if posterior:
            df_params = prettyifyParameterReport(df_params,variable,confidence)
            df_params = articulateParameters(df_params,axis=0)

        summ_path = assembleFullName(dir_path,'',file_name,'params','.txt')
        diux_path = assembleFullName(dir_path,'',file_name,'diauxie','.txt')

        #plate_cond.to_csv(file_path,sep='\t',header=True,index=True)
        df_params.to_csv(summ_path,sep='\t',header=True,index=posterior)
        if df_diauxie.shape[0]>0:
            df_diauxie.to_csv(diux_path,sep='\t',header=True,index=False)

        if save_latent:
            file_path = assembleFullName(dir_path,'',file_name,'output','.txt')
            x_out.to_csv(file_path,sep='\t',header=True,index=True)


    def computeFullDifference(self):
        '''
        Computes the full difference between two latent function (modelling growth curves).

        Args:
            x_diff (pandas.DataFrame): must include columns of Time, mu (mean of latent 
                function), Sigma (diagonal covariance of latent function)
            variable (str): variable of interest, must be a column name in x_diff
            confidence (float [0.0,1.0]): confidence interval, e.g. 0.95 for 95%.
            n (int): number of samples from posterior distribution
            posterior (boolean), whether to sample from posterior distribution
            noise (boolean): whether to plot 95-pct credibel intervals including sample uncertainty

        Returns:
            df (pandas.DataFrame)
            delta_od_sum (float): ||OD(t)||^2 which is defined as the sum of squares 
                for the OD when the mean and its credible interval deviates from zero.
        '''

        x_diff = self.x_full
        variable = self.target[0]
        confidence = self.args.confidence  # confidence interval, e.g. 0.95
        z_value = 1-(1 - confidence)/2
        noise = self.args.include_gaussian_noise
        save_latent = self.args.save_gp_data

        def buildTestMatrix(x_time):
            '''
            Build a test matrix to simlpify OD full difference computation.
                See https://github.com/ptonner/gp_growth_phenotype/testStatistic.py 
                This is used to compare two growth latent functions. The differeence between
                first time points (measurements) are adjusted to zero. 
            Args:
                x_time (pandas.DataFrame or pandas.Series or numpy.ndarray), ndim > 1
            Returns:
                A (numpy.ndarray): N-1 x 2*N where N is length of time.
            '''

            # buildtestmatrix
            n = x_time.shape[0]
            A = np.zeros((n-1,2*n))
            A[:,0] = 1
            A[range(n-1),range(1,n)] = -1
            A[:,n] = -1
            A[range(n-1),n+np.arange(1,n)] = 1

            return A

        x_diff = x_diff.sort_values([variable,'Time']) # do you really need to sort by variable
        x_time = x_diff.Time.drop_duplicates()

        # define mean and covariance of data
        mu = x_diff['mu'].values
        if noise:
            Sigma = np.diag(x_diff['Sigma'] + x_diff['Noise'])
        else:
            Sigma = np.diag(x_diff['Sigma'])

        # define mean and covariance of functional diffeence
        A = buildTestMatrix(x_time)
        m = np.dot(A,mu)
        c = np.dot(A,np.dot(Sigma,A.T))
        mean,std = m,np.sqrt(np.diag(c))

        # sample the curve for the difference between functions, from an MVN distribution
        n = getValue('n_posterior_samples')
        samples = np.random.multivariate_normal(m,c,n)
        
        # compute the sum of functional differences for all sampled curves
        dos = [np.sqrt(np.sum([ii**2 for ii in s])) for s in samples]
        dos_mu, dos_std = np.mean(dos), np.std(dos)

        # compute the confidence interval for the sum of functional differences
        scaler = norm.ppf(z_value) # define confidence interval scaler for MVN predictions
        ci = (dos_mu-scaler*dos_std, dos_mu+scaler*dos_std)

        # compute credible intervals for the curve of the difference
        y_avg = mean
        y_low = y_avg-scaler*std#
        y_upp = y_avg+scaler*std

        # package results
        t = x_time[1:].values
        df = pd.DataFrame([t,y_avg,y_low,y_upp],index=['Time','Avg','Low','Upp']).T

        self.functional_diff = df
        self.delta_od_sum_mean = dos_mu
        self.delta_od_sum_ci = ci
 
        # save gp_data fit
        dir_path = self.paths_dict['dir']
        file_name = self.paths_dict['filename']      
        if save_latent:
            file_path = assembleFullName(dir_path,'',file_name,'func_diff','.txt')
            df.to_csv(file_path,sep='\t',header=True,index=True)


    def plotPredictions(self):
        '''
        Visualizes the model tested by a specific hypothesis given the data.

        Args:
            x_full (pandas.DataFrame)
            x_min (pandas.DataFrame)
            hypotheis (dictionary): keys are str(H0) and str(H1), values are lists of str
            plate (growth.GrowthPlate obj))
            variable (list): variables of interest
            factor_dict (dictionary): mapping of unique values of variables to numerical integers
            subtract_control (boolean): where control sample curves subtracted from treatment sample curves
            file_name (str): 
            directory (str): path where files/figures should be stored
            args_dict (dictionary): must at least include 'nperm', 'nthin', and 'fdr' as keys and their values

        Action:
            saves a plot as PDF file
        '''

        # get necessary attributs
        x_full = self.x_full
        x_min = self.x_min
        factor_dict = self.factor_dict
        variable = self.target[0]
        plate = self.plate

        subtract_control = self.subtract_control
        directory = self.paths_dict['dir'] 
        file_name = self.paths_dict['filename']

        # get and modify user-accessible parameters from config.py
        plot_params = getHypoPlotParams()  # dict
        tick_spacing = plot_params['tick_spacing']
        legend_loc = plot_params['legend']
        fontsize = plot_params['fontsize']

        colors = getValue('hypo_colors')  # list of colors
        confidence = self.args.confidence  # confidence interval, e.g. 0.95
        z_value = 1-(1 - confidence)/2

        noise = self.args.include_gaussian_noise

        if self.args.dont_plot:
            return None

        # grab mapping of integer codes in design matrix to actual variable labels
        varb_codes_map = reverseDict(factor_dict[variable])  # {codes:vlaues}
 
        # set figure aesthetics
        sns.set_style('whitegrid')
        rcParams['font.family'] = 'sans-serif'
        rcParams['font.sans-serif'] = 'Arial'
  
        # initialize grid
        fig,ax = plt.subplots(2,1,figsize=[5,10.5],sharex=False)

        # for each unique value of variable of interest, plot MVN prediction
        list_values = varb_codes_map.items()
        list_values = sorted(list_values, key=operator.itemgetter(1))
        list_colors = colors[0:x_min.shape[0]]

        # plot MVN predictions
        for v_map,color in zip(list_values,list_colors):
            code,label = v_map
            criteria_real = {variable:[label]}
            criteria_mvn = {variable:code}

            ax[0] = addRealPlotLine(ax[0],plate,criteria_real,color,plot_params)
            ax[0] = addMVNPlotLine(ax[0],x_full,criteria_mvn,label,z_value,color,plot_params,noise)
            ax[0].xaxis.set_major_locator(MultipleLocator(tick_spacing))

        # adjust labels and window limits
        ax[0] = setAxesLabels(ax[0],subtract_control,plot_params,logged=self.args.log_transform)

        # if variable has only 2 values and if requested, plot delta OD
        if (len(list_values) != 2) or (self.args.dont_plot_delta_od):
            fig.delaxes(ax[1])
        else: 
            ax[1] = plotDeltaOD(ax[1],self.functional_diff,ylabel=True,xlabel=True,fontsize=fontsize)
            ax[1].xaxis.set_major_locator(MultipleLocator(tick_spacing))
            ax[0].set_xlabel('')

        ax = dynamicWindowAdjustment(ax)

        ## if user did not pass file name for output, use time stamp
        fig_path = assemblePath(directory,file_name,'.pdf')
        plt.subplots_adjust(wspace=0.15,hspace=0.15)
        savePlotWithLegends(ax[0],fig_path,legend_loc,fontsize=fontsize)


    def reportRegression(self):
        '''
        Describes the log Bayes Factor, and the percentile cut-offs for accepting H1 vs H0 based FDR <=20%.
            Results can be reported to stdout.

        Args:
            hypothesis (dictionary): e.g. {'H0':['Time'],'H1':['Time','Substrate']}
            log_BF (float): log Bayes Factor
            dist_log_BF (lsit of floats): log Bayes Factor based on permutation testing (i.e. null distribution)
            FDR (int): false discovery rate cuto-off (%)
            verbose (boolean)

        Returns:
            M1_Pct_Cutoff (float): FDR-based cut-off for accepting alt. model (actual BF must be higher)
            M0_Pct_Cutoff (flaot): FDR-based cut-off for accepting null model (actual BF must be lower)
            log_BF_Pct (float): percentile of actual log Bayes Factor relative to log Bayes Factor null distribution
            msg (str)
        '''

        hypothesis = self.hypothesis
        log_BF = self.log_BF
        dist_log_BF = self.log_BF_null_dist
        fdr = self.args.false_discovery_rate
        verbose = self.args.verbose

        log_BF_Display = prettyNumberDisplay(log_BF)
        dos_mean = prettyNumberDisplay(self.delta_od_sum_mean)
        dos_low = prettyNumberDisplay(self.delta_od_sum_ci[0])
        dos_upp = prettyNumberDisplay(self.delta_od_sum_ci[1])

        if dist_log_BF is None:

            msg = f'Model Tested: {hypothesis}\n\n' 
            msg += f'log Bayes Factor: {log_BF_Display}\n\n'
            msg += f'Functional Difference [95% CI]: {dos_mean} [{dos_low},{dos_upp}]\n'

            self.msg = f'\n{msg}'
            self.M1_Pct_Cutoff = None
            self.M0_Pct_Cutoff = None
            self.log_BF_Pct = None

            smartPrint(msg,verbose)

            return None 

        nperm = int(len(dist_log_BF)+1)

        # The 20% percentile in null distribution, a log BF higher has FDR <=20% that H1 fits data better than H0
        M1_Pct_Cutoff = np.percentile(dist_log_BF,100-fdr)
        M1_Display = prettyNumberDisplay(M1_Pct_Cutoff)

        # The 80% percentile in null distribution, a lo gBF lower has FDR <=20% that H0 fits data better than H1
        M0_Pct_Cutoff = np.percentile(dist_log_BF,fdr)
        M0_Display = prettyNumberDisplay(M0_Pct_Cutoff)

        # Percentile of actual log BF relative to null distribution
        log_BF_Pct = 100 - percentileofscore(dist_log_BF,log_BF) 
        
        msg = f'The following hypothesis was tested on the data:\n{hypothesis}\n\n' 
        msg += f'log Bayes Factor = {log_BF_Display} '
        msg += f'({log_BF_Pct:.1f}-percentile in null distribution based on {nperm} permutations)\n\n'
        msg += f'For P(H1|D) > P(H0|D) and FDR <= {fdr}%, log BF must be > {M1_Display}\n'
        msg += f'For P(H0|D) > P(H1|D) and FDR <= {fdr}%, log BF must be < {M0_Display}\n'
        msg += f'\nThe functional difference [95% CI] is {dos_mean} [{dos_low},{dos_upp}]\n'

        self.M1_Pct_Cutoff = M1_Pct_Cutoff
        self.M0_Pct_Cutoff = M0_Pct_Cutoff
        self.log_BF_Pct = log_BF_Pct
        self.msg = msg

        smartPrint(msg,verbose)


    def exportReport(self):

        def oneLineRport(**kwargs):
            return pd.DataFrame(columns=[0],index=list(kwargs.keys()),data=list(kwargs.values())).T

        if self.args.subtract_control:
            sc_msg = 'Samples were normalized to their respective control samples before analysis.'
        else:
            sc_msg = 'Samples were modeled without controlling for batch effects '
            sc_msg += '(i.e. subtracting the growth of group/batch-specific control samples).'

        nthin = len(np.unique(self.model.x[:,0]))

        msg = 'The following criteria were used to subset data:\n'
        msg += tidyDictPrint(self.params['subset'])
        msg += '\n'
        msg += self.msg
        msg += '\nData Manipulation: Input was reduced to '
        msg += f'{nthin} equidistant time points. {sc_msg}'
        self.msg = msg

        # compact report of results
        report_args = {'Filename':self.paths_dict['filename'],
                      'Subtract_Control':self.args.subtract_control,
                      'Subset':self.params['subset'],
                      'Hypothesis':self.params['hypothesis'],
                      'LL0':self.LL0,
                      'LL1':self.LL1,
                      'Log_BF':self.log_BF,
                      'FDR':self.args.false_discovery_rate,
                      'M1_FDR_cutoff':self.M1_Pct_Cutoff,
                      'M0_FDR_cutoff':self.M0_Pct_Cutoff,
                      'Permuted_log_BF':self.log_BF_null_dist,
                      'Func_Diff_Mean':self.delta_od_sum_mean,
                      'Func_Diff_CI':self.delta_od_sum_ci}

        dir_path = self.paths_dict['dir']
        file_name = self.paths_dict['filename']

        file_path = assembleFullName(dir_path,'',file_name,'log','.txt')
        self.compactReport = oneLineRport(**report_args)
        self.compactReport.to_csv(file_path,sep='\t',header=True,index=None)

        # save report of data
        file_path = assembleFullName(dir_path,'',file_name,'report','.txt')
        fid = open(file_path,'w')
        fid.write(self.msg)
        fid.close()



#!/usr/bin/env python

'''
AMiGA library of functions for trimming data and meta-data for AMiGA.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (6 functions)

# trimInput
# annotateMappings
# trimMergeMapping
# trimMergeData
# flagWells
# subsetWells

import copy
import pandas as pd # type: ignore
import sys

from functools import reduce

from .utils import subsetDf, resetNameIndex
from .comm import smartPrint, tidyDictPrint


def trimInput(data_dict,mapping_dict,params_dict=None,nskip=0,verbose=False):
    '''
    Interprets parameters to reduce mapping and
     data files to match user-desired criteria.

    Args:
        data (dictionary): keys are plate IDs and values are pandas.DataFrames with size t x (n+1)
            where t is the number of time-points and n is number of wells (i.e. samples),
            the additional 1 is due to the explicit 'Time' column, index is uninformative.
        mapping (dictionary): keys are plate IDs and values are pandas.DataFrames with size n x (p)
            where is the number of wells (or samples) in plate, and p are the number of variables or
            parameters described in dataframe.
        params (dictionary): must at least include 'subset' and 'flag' keys and their values
        verbose (boolean)

    Returns:
        data (dictionary): values may have smaller size than at time of input
        mapping (dictionary): values may have smaller size than at time of input 
    '''

    # a deep copy is needed to avoid altering mapping_dict globally
    mapping_dict = copy.deepcopy(mapping_dict)

    if params_dict is None:
        params_dict = {'subset': {}, 'flag': {}, 'hypothesis': {}, 'interval': {}}

    # annotate Subset and Flag columns in mapping files
    mapping_dict = annotateMappings(mapping_dict,params_dict,verbose)

    # trim and merge into single pandas.DataFrames
    master_mapping = trimMergeMapping(mapping_dict,verbose) # named index: Sample_ID
    master_data = trimMergeData(data_dict,master_mapping,nskip,verbose) # unnamed index: row number

    return master_data,master_mapping


def annotateMappings(mapping_dict,params_dict,verbose=False):
    '''
    Annotates the mapping data based on user-passed flags and subsetting criteria. In particular,
        it will annotate the Flag and Subset columns. It will also turn Well index into a
        standalone column. 

    Args:
        mapping_dict (dictionary): of mapping files (pandas.DataFrames), keys are file names (str)
        params_dict (dictionary): dictionary where keys are variables and values instances
        verbose (boolean)

    Returns:
        mapping_dict (dictionary): should be equal or smaller in size than input
    '''

    # flag wells that user does not want to analyze
    mapping_dict = flagWells(mapping_dict,params_dict['flag'],verbose=verbose,drop=False)

    # tag wells that meet user-passed criteria for analysis
    mapping_dict,_ = subsetWells(mapping_dict,params_dict['subset'],params_dict['hypothesis'],verbose=verbose)

    # make sure that mappings have Well columns
    #   here we assume that mapping_dict values have index of Well IDs, which should be the case
    mapping_dict = {pid:resetNameIndex(df,'Well',False) for pid,df in mapping_dict.items()}

    return mapping_dict


def trimMergeMapping(mapping_dict,verbose=False):
    '''
    Trims and merges mapping dataframes into one master mapping data frame.

    Args:
        mapping (dictionary): keys are plate IDs and values are pandas.DataFrames with size n x (p)
            where is the number of wells (or samples) in plate, and p are the number of variables or
            parameters described in dataframe.
        params (dictionary): must at least include 'subset' and 'flag' keys and their values
        verbose (boolean)

    Returns:
        mapping (pandas.DataFrame): number of wells/samples (n) x number of variables (p)  
    '''

    # merge mapping dataFrames
    #   sort will force shared (inner) keys to the lead and unshared (outer) keys to the caboose
    #   useful because individual mapping files may lack certain columns, some may even be empty 

    master_mapping = pd.concat(mapping_dict.values(),ignore_index=True,join='outer',sort=False)

    # trim mapping based on Subset and Flag columns
    master_mapping = subsetDf(master_mapping,{'Subset':[1],'Flag':[0,1]})

    # reset_index and set as Sample_ID
    master_mapping = resetNameIndex(master_mapping,'Sample_ID',True)

    return master_mapping


def trimMergeData(data_dict,master_mapping,nskip=0,verbose=False):
    '''
    Trims and merges data dataframes into one master dataframe. 

    Args:
        data (dictionary): keys are plate IDs and values are pandas.DataFrames with size t x (n+1)
            where t is the number of time-points and n is number of wells (i.e. samples),
            the additional 1 is due to the explicit 'Time' column, index is uninformative.
        mapping_df (pandas.DataFrame): number of well/samples (n) x number of variables (p)
        verbose (boolean)

    Returns:
        master_data (pandas.DataFrame): number of time points (t) x number of variables plus-one (p+1)
            plus-one because Time is not an index but rather a column
    '''

    new_data_dict = {}

    for pid,data in data_dict.items():

        # grab plate-specific samples
        mapping_df = master_mapping[master_mapping.Plate_ID==pid]
        
        # if no plate-specific samples are included in master_mapping, skip
        if mapping_df.shape[0]==0:
            continue

        # grab plate-specific data
        wells = list(mapping_df.Well.values)
        df = data.loc[:,['Time']+wells]

        # update plate-specific data with unique Sample Identifiers 
        sample_ids = list(mapping_df.index.values)
        df.columns = ['Time'] + sample_ids
        df.T.index.name = 'Sample_ID'

        # udpate dictionary
        df = df.iloc[nskip:,:]
        new_data_dict[pid] = df

    # each data is time (T) x wells (N), will merge all data and keep a single Time column
    #     reduce(fun,seq) applies a function (fun) recursively to all elements in list (seq)
    #     here, reduce will merge the first two dataFrames in data_dict.values(), then it
    #     will take this output and merge it with third dataFrame in list, and so on
    
    try: 
        master_data = reduce(lambda ll,rr: pd.merge(ll,rr,on='Time',how='outer'),new_data_dict.values())
    except TypeError:
        msg = "\nFATAL USER ERROR: AMiGA could not subset data based on user input. "
        msg += "Please check your arguments especially '-s' or '--subset' for any typos. "
        msg += "Keep in mind that AMiGA argument parser is case-sensitive.\n"
        sys.exit(msg)

    master_data = master_data.sort_values(['Time']).reset_index(drop=True)
    
    return master_data
    

def flagWells(df,flags,verbose=False,drop=False):
    '''
    Passes plate-well-specific flags from user into mapping dataframes.

    Args:
        df (dictionary of pandas.DataFrame) must have Plate_IDs and Well as columns
        flags (dictionary) with Plate_IDs (str) as keys and Wells (stR) as vlaues
        verbose (boolean)

    Returns:
        df (pandas.DataFrame)
    '''

    if (len(flags)==0):
        smartPrint('No wells were flagged.\n',verbose)
        return df

    for plate, wells in flags.items():
        if plate in df.keys():

            df[plate].loc[wells,'Flag'] = [1]*len(wells)

            if drop:
                df[plate] = df[plate][df[plate].Flag==0] 

    smartPrint('The following flags were detected:\n',verbose)
    smartPrint(tidyDictPrint(flags),verbose)

    return df


def subsetWells(df_mapping_dict,criteria,hypothesis,verbose=False):
    '''
    Tag wells that meet user-passed criteria.

    Args:
        df (pandas.DataFrame) must have Plate_IDs and Well as columns
        criteria (dictionary) with mapping variables (str) as keys and accepted instances (str) as values
        hypothesis (dictionary) 
        verbose (boolean)

    Returns:
        df (pandas.DataFrame)
    '''

    if (len(criteria)==0):
        smartPrint('No subsetting was requested.\n',verbose)
        return df_mapping_dict,None

    for plate_id,mapping_df in df_mapping_dict.items():

        # subsetting on number-based criteria does not match hits due to type mismatch (str vs int/float)
        mapping_df_str = mapping_df.astype(str)

        remove_boolean = ~(mapping_df_str.isin(criteria).sum(axis=1,numeric_only=True)==len(criteria)).values  # list of booleans
        remove_idx = mapping_df_str.index[remove_boolean]
        mapping_df.loc[remove_idx,'Subset'] = [0]*len(remove_idx)

    msg = 'The following criteria were used to subset data:\n'
    msg += tidyDictPrint(criteria)

    smartPrint(msg,verbose)

    return df_mapping_dict,msg




#!/usr/bin/env python

'''
AMiGA library of functions for processing and detailing meta-data.
'''

__author__ = "Firas Said Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (18 functions)

# assembleMappings
# checkBiologSize
# checkMetaText
# checkPlateIdColumn
# isBiologFromMeta
# grabFirstValueFromDf
# expandBiologMetaData
# initSubstrateDf
# isBiologFromName
# initKeyFromMeta
# initBiologPlateKey
# parsePlateName
# initMappingDf
# expandMappingParams
# parseBiologLayout
# parseWellLayout
# updateMappingControls
# shouldYouSubtractControl

import os 
import sys
import numpy as np
import pandas as pd

from string import ascii_uppercase
from tabulate import tabulate

from . import biolog
from .comm import smartPrint
from .org import assembleFullName, assemblePath
from .utils import subsetDf


def assembleMappings(data,mapping_path,meta_path=None,save=False,verbose=False):
    '''
    Creates a master mapping file (or dictionary ?) for all data files in the input argument.
        For each data file, in this particular order, it will first (1) check if an individual
        mapping file exists, (2) if not, check if relevant meta-data is provided in meta.txt
        file, (3) if not, infer if plate is a BIOLOG PM based on its file name, and (4) if all
        fail, create a minimalist mapping file. 

    Args:
        data (dictionary): keys are file names (i.e. filebases or Plate IDs) and values are
            pandas DataFrames where index column (row names) are well IDs.
        mapping_path (str): path to the mapping folder.
        meta_path (str): path to the mapping file.
        verbose (boolean)

    Returns:
        df_mapping_dict (dictionary): keys are file names and values are mapping files. 
    '''

    df_mapping_dict = {}

    # list all data files to be analyed
    list_filebases = data.keys()

    # list all potential mapping file paths
    list_mapping_files = [assemblePath(mapping_path,ii,'.txt') for ii in list_filebases]

    # read meta.txt and list all plates described by it
    meta_df, meta_df_plates = checkMetaText(meta_path,verbose=verbose)   

    # assemble mapping for one data file at a time
    for filebase,mapping_file_path in zip(list_filebases,list_mapping_files):

        # what are the row names from the original data file 
        well_ids = data[filebase].columns[1:]  # this may no be A1 ... H12, but most ofen will be

        # create file path for saving derived mapping, if requested
        newfilepath = assembleFullName(mapping_path,'',filebase,'','.map')

        # see if user provided a mapping file that corresponds to this data file (filebase)
        if os.path.exists(mapping_file_path):

            df_mapping = pd.read_csv(mapping_file_path,sep='\t',header=0,index_col=0, dtype={'Plate_ID':str,'Isolate':str})
            df_mapping = checkPlateIdColumn(df_mapping,filebase) # makes sure Plate_ID is a column
            df_mapping.index = [ii[0] + ii[1:].lstrip('0') for ii in df_mapping.index] # strip leading zeros in well names

            smartPrint(f'{filebase:.<30} Reading {mapping_file_path}.',verbose=verbose)
        
        # see if user described the file in meta.txt 
        elif filebase in meta_df_plates:

            meta_info = meta_df[meta_df.Plate_ID==filebase]
            msg = f'{filebase:.<30} Found meta-data in meta.txt '

            biolog = isBiologFromMeta(meta_info)  # does meta_df indicate this is a BIOLOG plate

            if biolog:
                checkBiologSize(data[filebase],filebase)
                df_mapping = expandBiologMetaData(meta_info)
                msg += '& seems to be a BIOLOG PM plate.'
                smartPrint(msg,verbose=verbose)
            else:
                df_mapping = initKeyFromMeta(meta_info,well_ids)
                msg += '& does not seem to be a BIOLOG PM plate.'
                smartPrint(msg,verbose=verbose)

        elif isBiologFromName(filebase):
            checkBiologSize(data[filebase],filebase)
            df_mapping = initBiologPlateKey(filebase)
            msg = f'{filebase:.<30} Did not find mapping file or meta-data '
            msg += 'BUT seems to be a BIOLOG PM plate.'
            smartPrint(msg,verbose=verbose)

        else:
            df_mapping = initMappingDf(filebase,well_ids) 
            msg = f'{filebase:.<30} Did not find mapping file or meta-data '
            msg += '& does not seem to be a BIOLOG PM plate.'
            smartPrint(msg,verbose=verbose)

        df_mapping_dict[filebase] = expandMappingParams(df_mapping,verbose=verbose)

        if save:
            df_mapping_dict[filebase].to_csv(newfilepath,sep='\t',header=True,index=True)

    #df_mapping = df_mapping.reset_index(drop=False)
    smartPrint('',verbose=verbose)

    return df_mapping_dict


def checkBiologSize(df,filename):
    '''
    If a Plate_ID or meta.txt suggest a file corresponds to a BIOLOG PM, double check
        that it is at least a 96-well plate. 
    '''
    if df.shape[1] != 97: # (96 wells + 1 time) = 97 columns

        msg = f'The file "{filename}" is not formatted as a 96-well plate. '
        msg += 'Either the file name or its information in meta.txt '
        msg += 'suggsts that it is a BIOLOG PM plate. Please either correct the name '
        msg += 'of the file, the contents of the file, or correct its meta-data. '
        msg += 'If your file is missing Well IDs, you will need to add them in the first column. '
        msg += 'See documentation for more details.'

        sys.exit(msg)


def checkMetaText(filepath,verbose=False):
    '''
    Parses meta.txt file into a pandas.DataFrame.

    Args:
        filepath (str): path to meta.txt, must be tab-delimited with first column as "Plate_ID" (i.e. file name)
        verbose (boolean)

    Returns:
        df_meta (pandas.DataFrame)
        df_meta_plates (list)

    '''

    if filepath is None:
        exists = False
    else:
        exists = os.path.exists(filepath)

    if not exists:
        df_meta = pd.DataFrame
        df_meta_plates = []
    else:
        df_meta = pd.read_csv(filepath,sep='\t',header=0,index_col=None,dtype={'Plate_ID':str,'Isolate':str})

    # which plates were characterized in meta.txt?
    try:
        df_meta_plates = df_meta.Plate_ID.values
    except Exception:
        df_meta_plates = []

    # neatly prints meta.txt to terminal
    if exists:
        tab = tabulate(df_meta,headers='keys',tablefmt='psql')
        msg = '{:.<21}{}\n{}\n'.format('Meta-Data file is',filepath,tab)
        smartPrint(msg,verbose)
    else:
        smartPrint('No meta.txt file found\n',verbose)

    return df_meta,df_meta_plates


def checkPlateIdColumn(df,filebase):
    '''
    Validates that Plate ID is a column in dataframe.

    Args:
        df (pandas.DatamFrame)

    Returns:
        df (pandas.DataFrame) 
    '''

    if 'Plate_ID' not in df.keys():
        df.loc[:,'Plate_ID'] = [filebase]*df.shape[0]

    return df


def isBiologFromMeta(series):
    '''
    Identifies if a data file is a Biolog plate based on its meta-data.

    Args:
        ser (pandas.Series)

    Returns:
        (boolean)
    '''

    if 'PM' not in series.keys():
        return False
    elif grabFirstValueFromDf(series,'PM') in range(1,7): 
        return True
    else:
        return False


def grabFirstValueFromDf(df,key,fillna=None):
    '''
    Grabs the first value in a specific column in either a data frme or a series.

    Args:
        df (pandas.DataFrame or pandas.Series)
        key (str): column key
        fillna (str or None): if value not found, should replace with this argument

    Returns:
        value (data type is flexible)
    '''

    if key in df.keys():
        return df.loc[:,key].iloc[0]
    else:
        return fillna


def expandBiologMetaData(sr):
    '''
    Builds a pandas.DataFrame that combines file-specific meta-data with info derived from
        BIOLOG PM substrate lay-out. Datafrmae will include columns for Plate ID, Isolate 
        name, PM number, and Replicate number. 

    Args:
        sr (pandas.Series): describes meta-data for a specific file

    Returns:
        df (pandas.DataFrame)
    '''

    pmn = grabFirstValueFromDf(sr,'PM')  # PM plate number
    df_substrates = initSubstrateDf(pmn)  # mapping of wells to substrates
    df_meta = initKeyFromMeta(sr,df_substrates.index)
    df = df_meta.join(df_substrates)

    return df


def initSubstrateDf(pmn):
    '''
    Initializes a pandas DataFrame for substrates corresponding to a specific BIOLOG PM plate.

    Args:
        pmn (str or int or float): Phenotypic Microarray plate number, must be between [1,7)
    '''

    sr = parseBiologLayout().loc[:,str(int(pmn))]
    df = pd.DataFrame(sr)
    df.columns = ['Substrate']

    return df


def isBiologFromName(filebase):
    '''
    Determines if file name nomenclature indicates Biolog PM plate. Biolog PM nomenclature is
        {Isolate Name}_PM{PM plate number}-{Replicate number}.

    Args:
        filebase (str)

    Returns:
        (boolean) 
    '''

    if '_PM' in filebase:
        split_filebase = filebase.split('_PM')
    else:
        return False

    if split_filebase[1][0] not in [str(ii) for ii in range(1,7)]:
        return False

    return True


def initKeyFromMeta(series,well_ids):
    '''
    Initializes a mapping data frame using data derived from meta.txt. Here, it will reproduce
        a pandas.Series to match the number of wells in corresponding data file.

    Args:
        sr (pandas.Series): no size limit
        well_ids (list): list of variable data type

    Returns:
        df_meta (pandas.DataFrame): # rows = len(well_ids), # cols = df.shape[1]
    '''

    df_meta = pd.concat([series]*len(well_ids))
    df_meta.index = well_ids
    df_meta.index.name = 'Well'

    return df_meta


def initBiologPlateKey(plate_id,simple=False):
    '''
    Initiailzes a mapping dataframe using only Plate ID. Dataframe will have columns
        for Plate_ID, Isolate name, PM number, and Replicate number.

    Args:
        plate_id (str): data file name (i.e. filebase)
        simple (boolean): whether to include replicate number in mapping dataframe or not

    Returns:
        df_mapping (pandas.DataFrame)
    '''

    list_keys = ['Plate_ID','Isolate','PM','Replicate']
    isolate,pmn,rep = parsePlateName(plate_id,simple = False)
    df_meta = pd.DataFrame(index=list_keys,data=[plate_id,isolate,pmn,rep]).T
    df = expandBiologMetaData(df_meta)

    return df


def parsePlateName(plate_id,simple=False):
    '''
    Deconstructs a file name to identify Isolate name, PM plate number, and Replicate number.

    Args:
        plate_id (str)
        simple (boolean): whether to return Replicate number

    Returns:
        isolate (str)
        pmn (int)
        rep (int)
    '''

    plate_id_split = plate_id.split('_PM') 

    # in case isolate name includes "_PM" we will join all split strings except for last
    isolate = "_PM".join(plate_id_split[0:-1])

    # the number immediately fllowing the last instance of "_PM"
    pmn = int(plate_id_split[-1][0])
    
    # the number immediately following a hyphen after the last instance of "_PM", otherwise 1
    rep = [int(plate_id_split[-1].split('-')[-1]) if '-' in plate_id else 1][0]

    if simple:
        return isolate,pmn
    else:
        return isolate,pmn,rep


def initMappingDf(filebase,well_ids):
    '''
    Creates a minimalist file mapping (pandas.DataFrame). Index column are Well IDs, and sole
        value column is 'Plate_ID'. 

    Args:
        filebase (str): Plate ID
        well_ids (list of str): list of Well IDs (e.g. A1,A2,...,H11,H12)

    Returns:
        df_mapping (pandas.DataFrame): n x 1 where n is based on length of well_ids argument.
    '''

    data = [filebase]*len(well_ids)

    df_mapping = pd.DataFrame(index=well_ids,columns=['Plate_ID'],data=data)
    df_mapping.index.name = 'Well'

    return df_mapping


def expandMappingParams(df,verbose):
    '''
    Expand input data frame to include columns relevant for AMiGA processing of user paramters.
        It will add columns for Group, Control, Flag, and Subset. Note on grouping: plates
        can be divided into multiple groups where each group has its own group-specific 
        control wells. Biolog PM plates has a single group and A1 is control well. 

    Args:
        df (pandas.DataFrame)
        verbose (boolean)

    Returns:
        df (pandas.DataFrame): with four additional columns
    '''

    # get dataframe info
    Plate_ID = Plate_ID = df.Plate_ID.unique()[0]
    keys = list(df.keys()) 
    
    # check if Biolog PM plate
    biolog = isBiologFromName(Plate_ID) or isBiologFromMeta(df) # True or False

    if ('Control' in keys) and ('Group' not in keys):
        
        df.loc[:,'Control'] = df.Control.fillna(0)
        df.loc[:,'Group'] = [1]*df.shape[0]
                
    if ('Group' in keys) and ('Control' not in keys):

        df.loc[:,'Group'] = df.Group.fillna('NA')
        df.loc[:,'Control'] = [0]*df.shape[0]        
    
    if ('Group' not in keys) and ('Control' not in keys): 
    
        # plate can be divided into multiple group, each gorup has unique control well(s)
        df.loc[:,'Group'] = [1]*df.shape[0]  # all wells in a BIOLOG PM plte belong to same group
        df.loc[:,'Control'] = [0]*df.shape[0]  # all wells (except) A1 are treatments
    
    if biolog:      

        df.loc[:,'Control'] = 0  # A1 is the control well
        df.loc['A1','Control'] = 1  # A1 is the control well

    if not all(x in [0.,1.] or np.isnan(x) for x in df.Control.unique()):
        
        msg = '\nUSER ERROR: Values in Control column for mapping '
        msg += f'of {Plate_ID} must be either 0 or 1.\n'
        smartPrint(msg,verbose)

        df.loc[:,'Control'] = [0]*df.shape[0]

    # replace na values
    df.loc[:,'Group'] = df.Group.fillna('NA')
    df.loc[:,'Control'] = df.Control.fillna(0)

    # initialize well-specific flag and subset parameters
    if 'Flag' not in df.keys():
        df.loc[:,'Flag'] = [0]*df.shape[0]  # by default, no wells are flagged
    if 'Subset' not in df.keys():
        df.loc[:,'Subset'] = [1]*df.shape[0]  # by default, all wells are included in analysis

    return df


def parseBiologLayout():
    '''
    Initializes a pandas.DataFrame that maps the location (well and plate number)
        of each sugar in BIOLOG PM plates.

    Retruns:
        df (pandas.DataFrame)
    '''

    biolog_layout = [biolog.Carbon1,biolog.Carbon2,biolog.Nitrogen,biolog.PhosphorusAndSulfur]
    biolog_layout += [biolog.PeptideNitrogen1,biolog.PeptideNitrogen2,biolog.PeptideNitrogen3]

    index = [str(ii) for ii in range(1,8)]
    keys = parseWellLayout(order_axis=0).index

    biolog_layout_df = pd.DataFrame(biolog_layout,index=index,columns=keys).T
    biolog_layout_df.columns.name = 'PM'

    return biolog_layout_df 


def parseWellLayout(order_axis=1):
    '''
    Initializes a pandas.DataFrame where indices (row names) are well IDs (e.g. C8) and
        variables indicate row letter and column number.

    Args:
        order_axis (int): 0 indicates order by row (i.e. A1,A2,A3,...,B1,B2,B3,...) and
            1 indicates order by column (i.e. A1,B1,C1,...,A2,B2,C2,...).

    Returns:
        df (pandas.DataFrame, size 96x2): row names are well ID, and columns include
            row letter (str), and column number (str)
    '''

    # initialize rows = ['A',...,'H'] and cols = [1,...,12]
    rows = list(ascii_uppercase[0:8])
    cols = list(int(ii) for ii in range(1,13))

    list_wells = []
    list_cols = []
    list_rows = []

    # append one well at a time to list tht will later be merged into dataframe
    if order_axis == 1:
        for col in cols:
            for row in rows:
                list_wells.append(f'{row}{col}')
                list_rows.append(row)
                list_cols.append(col)
    else:
        for row in rows:
            for col in cols:
                list_wells.append(f'{row}{col}')
                list_rows.append(row)
                list_cols.append(col)

    # assemble dataframe
    df = pd.DataFrame([list_wells,list_rows,list_cols],index=['Well','Row','Column'])
    df = df.T  # transpose to long format 
    df = df.set_index('Well')

    return df


def updateMappingControls(master_mapping,mapping_dict,to_do=False):
    '''
    For all samples in master mapping, find relevant controls and add these controls to the master mapping dataframe.

    Args:
        master_mapping (pandas.DataFrame)
        mapping_dict (dictionary)
        to_do (boolean)

    Returns:
        master_mapping (pandas.DataFrame): will have more rows (i.e. samples) than input
    '''

    # check first if you need to do this
    if not to_do:
        return master_mapping

    # find all unique groups
    plate_groups = master_mapping.loc[:,['Plate_ID','Group']].drop_duplicates()
    plate_groups = [tuple(x) for x in plate_groups.values]

    # grab all relevant control samples
    df_controls = []
    for plate_group in plate_groups:
        pid,group = plate_group
        pid_mapping = mapping_dict[pid]
        df_controls.append(subsetDf(pid_mapping,{'Plate_ID':[pid],'Group':[group],'Control':[1]}))

    # re-assemble the master mapping dataframe, including the propercontrols
    df_controls = pd.concat(df_controls)
    master_mapping = pd.concat([master_mapping.copy(),df_controls.copy()],sort=True)
    master_mapping = master_mapping.reset_index(drop=True)
    master_mapping.index.name = 'Sample_ID'
    master_mapping = master_mapping.sort_values(['Plate_ID','Group','Control'])

    # if mapping has an interaction column, replace NaN with 0 (so it won't be dropped later)
    #   because you are (above) adding control samples to master_mapping,
    #   they will not have the interaction column and their values will default to NaN
    variable = [v for v in master_mapping.keys() if '*' in v]
    master_mapping.loc[:,variable] = master_mapping.loc[:,variable].fillna(0)

    return master_mapping


def shouldYouSubtractControl(mapping,variables):
    '''
    Checks if control samples must be subtracted from treatment samples for proper hypothesis testing.
        In particular, make sure that the variable of interest is binary (i.e. it has only two possible
        values in the mapping dataframe. This makes sure that GP regression on variable of interest is 
        performing a test on a binary variable.

    Args:
        mapping (pandas.DataFrame): samples (n) by variables (k)
        variable (str): must be one of the column headers for mapping argument

    Returns:
        (boolean)
    '''

    unique_values = mapping.loc[:,variables].drop_duplicates().reset_index()

    for _,row in unique_values.iterrows():
        criteria = row.to_dict()
        sub_map = subsetDf(mapping,criteria)
        sub_map_controls_n = sub_map[sub_map.Control==1].shape[0]
        sub_map_total_n = sub_map.shape[0]
        if (sub_map_controls_n == sub_map_total_n) and (sub_map_controls_n > 0):
            return False  # found a value whose samples all correspond to control wells
    else:
        return True
    


#!/usr/bin/env python

'''
AMiGA wrapper: comparing parameters between two user-defined conditions.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (6 functions)

# main
# save
# read
# validate
# subset
# compare


import sys
import pandas as pd # type: ignore

from .interface import checkParameterCommand
from .params import initParamList, articulateParameters
from .utils import subsetDf


def main(args):

	if (args.confidence < 80) | (args.confidence > 100):
		msg = '\nFATAL USER ERROR: Confdience must be between 80 and 100.\n'
		sys.exit(msg) 

	df = validate(read(args))
	df,varbs = subset(args,df)
	df = compare(args,df,varbs)
	save(args,df)


def save(args,df):

	df = articulateParameters(df,axis=0)
	df.to_csv(f'{args.output}.txt',sep='\t',header=False,index=True)


def read(args):

	ls_df = []

	for ii in args.input:
		ls_df.append(pd.read_csv(ii,sep='\t',header=0,index_col=None))

	df = pd.concat(ls_df,sort=False).reset_index(drop=True)

	return df


def validate(df):
	'''
	Do the columns include summary statistics for growth parameters? Look for 
	column headers that have the format of mean(...) and std(...).

	Args:
		df (pandas.DataFrame)

	Returns: 
		(boolean): True or False
	'''
	cols = df.keys()
	means = [True if ii.startswith('mean') else False for ii in cols]
	stds = [True if ii.startswith('std') else False for ii in cols]
	if (not any(means)) | (not any(stds)):
		msg = '\nFATAL USER ERROR: The input file does not contain summary '
		msg += 'statistics for growth parameter, such as mean and standar deviation.\n'
		sys.exit(msg)
	else:
		return df


def subset(args,df):

	ls_df, ls_varbs = [], []

	if isinstance(args.subset,str):
		args.subset = [args.subset]

	for ii in args.subset:
		criteria = checkParameterCommand(ii,sep=',')
		ls_df.append(subsetDf(df,criteria))
		ls_varbs.append(list(criteria.keys()))

	df = pd.concat(ls_df,sort=False).reset_index(drop=True).drop_duplicates()

	if df.shape[0] != 2:

		msg = '\nFATAL USER ERROR: User-provided summary files and subsetting crieteria '
		msg += 'selected for more or less than two conditions. AMiGA can only perform '
		msg += 'comparison on two conditions. Please check your arguments and try again. '
		msg += 'Below are the currently selected conditions.\n\n'

		keys = [ii for ii in df.keys() if ('(' not in ii) & (ii != 'diauxie') ]

		print(msg)
		print(df.loc[:,keys],'\n\n')
		sys.exit()

	else:

		ls_varbs = [item for sublist in ls_varbs for item in sublist]
		ls_varbs = list(set(ls_varbs))

		return df,ls_varbs


def compare(args,df,varbs):


	def getConfInts(means,stds,z_value=0.975):
		'''
		Computes confidence interval based on mean, standard deviation, and desired confidence.

		Args:
			means (array of floats)
			stds (array of floats)
			z_value (float)

		Returns:
			cis (array of strings), where each formatted string indicates the confidence interval,
			e.g. [0.4,0.6]
		'''

		from scipy.stats import norm

		scaler = norm.ppf(z_value)

		cis = []
		for m,s in zip(means,stds):
			low, upp = m-scaler*s, m+scaler*s
			cis.append(f'[{low:.3f},{upp:.3f}]')

		return cis

	def detSigDiff(a,b):
		'''
		Detemines if there is a significant difference between two variables, 
		based on whether confidence intervals overlap or not

		Args:
			a (2-array): lower and upper bounds of first confidence interval
			b (2-array): lower and upper bounds of second confidence intervals

		Returns:
			(boolean): True (intervals do not overlap) or False (intervals overlap)
		'''

		a = [float(ii) for ii in a]
		b = [float(ii) for ii in b]

		return not ((a[0] <= b[1]) and (b[0] <= a[1]))

	z_value = (100 - (100 - args.confidence)/2) / 100 

	if 'Sample_ID' in df.keys():
		df = df.set_index(['Sample_ID'])

	params = set(df.keys()).difference(set(varbs))
	params = list(params.intersection(initParamList(1)+['diauxie']))
	params = list({ii.split('(')[1][:-1] if '(' in ii else ii for ii in params})

	df_top = df.loc[:,varbs].reset_index(drop=True).T

	df_bot = pd.DataFrame(index=params,columns=df_top.columns).sort_index()
	df_mus = pd.concat([df_top,df_bot])

	df_bot = pd.DataFrame(index=params,columns=list(df_top.columns)+['']).sort_index()
	df_cis = pd.concat([df_top,df_bot])

	for p in params:
		if p == 'diauxie':
			df_mus.loc[p,:] = df.loc[:,p].values
			if df.loc[:,p].values[0] != df.loc[:,p].values[1]:
				df_cis.loc[p,:] = ['NA','NA',True]
			else:
				df_cis.loc[p,:] = ['NA','NA',False]
		else:
			mus = df.loc[:,f'mean({p})'].values
			stds = df.loc[:,f'std({p})'].values
			cis = getConfInts(mus,stds,z_value)
			olap = detSigDiff(eval(cis[0]),eval(cis[1]))

			df_mus.loc[p,:] = [f'{ii:.3f}' for ii in mus]
			df_cis.loc[p,:] = cis + [olap]

	df_mus.loc['Parameter',:] = ['Mean','Mean']
	df_cis.loc['Parameter',:] = [f'{args.confidence}% CI',
	                             f'{args.confidence}% CI',
	                             'Sig. Diff.']

	df_all = df_mus.join(df_cis,lsuffix='L',rsuffix='R')
	df_all = df_all.loc[varbs+['Parameter']+sorted(params),:]
	df_all = df_all.T.reset_index(drop=True).T

	return df_all

